# ruff: noqa: D100 D103
import warnings


def test_import():
    with warnings.catch_warnings():
        warnings.simplefilter(action="ignore", category=FutureWarning)
        import pastastore  # noqa: F401



# ruff: noqa: D100 D103
import numpy as np
import pandas as pd
import pytest
from conftest import requires_pkg

import pastastore as pst

# %% write

# data
rng = np.random.default_rng()
data = rng.random(int(1e5))
s = pd.Series(index=pd.date_range("1970", periods=int(1e5), freq="h"), data=data)
metadata = {"x": 100000.0, "y": 300000.0}


def series_write(conn):
    conn.add_oseries(s, "oseries1", metadata=metadata, overwrite=True)


# @pytest.mark.benchmark(group="write_series")
# def test_benchmark_write_series_dict(benchmark):
#     conn = pst.DictConnector("test")
#     _ = benchmark(series_write, conn=conn)


@pytest.mark.benchmark(group="write_series")
def test_benchmark_write_series_pas(benchmark):
    conn = pst.PasConnector("test", "./tests/data/pas")
    _ = benchmark(series_write, conn=conn)


@pytest.mark.benchmark(group="write_series")
@requires_pkg("arcticdb")
def test_benchmark_write_series_arcticdb(benchmark):
    uri = "lmdb://./arctic_db/"
    conn = pst.ArcticDBConnector("test", uri)
    _ = benchmark(series_write, conn=conn)


# %% read


def series_read(conn):
    _ = conn.get_oseries("oseries1")


# @pytest.mark.benchmark(group="read_series")
# def test_benchmark_write_series_dict(benchmark):
#     conn = pst.DictConnector("test")
#     _ = benchmark(series_read, conn=conn)
#


@pytest.mark.benchmark(group="read_series")
def test_benchmark_read_series_pas(benchmark):
    conn = pst.PasConnector("test", "./tests/data/pas")
    _ = benchmark(series_read, conn=conn)


@pytest.mark.benchmark(group="read_series")
@requires_pkg("arcticdb")
def test_benchmark_read_series_arcticdb(benchmark):
    uri = "lmdb://./arctic_db/"
    conn = pst.ArcticDBConnector("test", uri)
    _ = benchmark(series_read, conn=conn)


# %% write model


def build_model(conn):
    store = pst.PastaStore(conn, "test")

    # oseries nb1
    if "oseries_nb1" not in store.oseries_names:
        o = pd.read_csv("./tests/data/head_nb1.csv", index_col=0, parse_dates=True)
        store.add_oseries(o, "oseries_nb1", metadata={"x": 100300, "y": 400400})

    # prec nb1
    if "prec_nb1" not in store.stresses.index:
        s = pd.read_csv("./tests/data/rain_nb1.csv", index_col=0, parse_dates=True)
        store.add_stress(
            s, "prec_nb1", kind="prec", metadata={"x": 100300, "y": 400400}
        )

    # evap nb1
    if "evap_nb1" not in store.stresses.index:
        s = pd.read_csv("./tests/data/evap_nb1.csv", index_col=0, parse_dates=True)
        store.add_stress(
            s, "evap_nb1", kind="evap", metadata={"x": 100300, "y": 400400}
        )

    ml = store.create_model("oseries_nb1", add_recharge=True)

    return ml


def write_model(conn, ml):
    conn.add_model(ml, overwrite=True)


# @pytest.mark.benchmark(group="write_model")
# def test_benchmark_write_model_dict(benchmark):
#     conn = pst.DictConnector("test")
#     ml = build_model(conn)
#     _ = benchmark(write_model, conn=conn, ml=ml)


@pytest.mark.benchmark(group="write_model")
def test_benchmark_write_model_pas(benchmark):
    conn = pst.PasConnector("test", "./tests/data/pas")
    ml = build_model(conn)
    _ = benchmark(write_model, conn=conn, ml=ml)


@pytest.mark.benchmark(group="write_model")
@requires_pkg("arcticdb")
def test_benchmark_write_model_arcticdb(benchmark):
    uri = "lmdb://./arctic_db/"
    conn = pst.ArcticDBConnector("test", uri)
    ml = build_model(conn)
    _ = benchmark(write_model, conn=conn, ml=ml)


# %%


def write_model_nocheckts(conn, ml):
    conn.set_check_model_series_values(False)
    conn.add_model(ml, overwrite=True)


@pytest.mark.benchmark(group="write_model")
def test_benchmark_write_model_nocheckts_pas(benchmark):
    conn = pst.PasConnector("test", "./tests/data/pas")
    ml = build_model(conn)
    _ = benchmark(write_model_nocheckts, conn=conn, ml=ml)


@pytest.mark.benchmark(group="write_model")
@requires_pkg("arcticdb")
def test_benchmark_write_model_nocheckts_arcticdb(benchmark):
    uri = "lmdb://./arctic_db/"
    conn = pst.ArcticDBConnector("test", uri)
    ml = build_model(conn)
    _ = benchmark(write_model_nocheckts, conn=conn, ml=ml)


# %% read model


def read_model(conn):
    ml = conn.get_models("oseries_nb1")
    return ml


# @pytest.mark.benchmark(group="read_model")
# def test_benchmark_read_model_dict(benchmark):
#     conn = pst.DictConnector("test")
#     _ = benchmark(read_model, conn=conn, ml=ml)


@pytest.mark.benchmark(group="read_model")
def test_benchmark_read_model_pas(benchmark):
    conn = pst.PasConnector("test", "./tests/data/pas")
    _ = benchmark(read_model, conn=conn)
    pst.util.delete_pas_connector(conn)


@pytest.mark.benchmark(group="read_model")
@requires_pkg("arcticdb")
def test_benchmark_read_model_arcticdb(benchmark):
    uri = "lmdb://./arctic_db/"
    conn = pst.ArcticDBConnector("test", uri)
    _ = benchmark(read_model, conn=conn)
    pst.util.delete_arcticdb_connector(conn=conn)
    import shutil

    shutil.rmtree("./arctic_db/")



# ruff: noqa: D100 D103
import pastas as ps
import pytest


def test_stressmodel_time_series_name(pstore):
    pstore.get_stressmodel("evap1")


def test_stressmodel_override_settings(pstore):
    sm = pstore.get_stressmodel("prec1", settings="evap")
    assert sm.stress[0].settings["fill_nan"] == "interpolate"

    # override settings by passing kind: bit weird, but good to check
    sm = pstore.get_stressmodel("prec1", kind="evap")
    assert sm.stress[0].settings["fill_nan"] == "interpolate"


def test_stressmodel_rfunc_kwargs(pstore):
    sm = pstore.get_stressmodel("well1", rfunc=ps.Hantush, rfunc_kwargs={"quad": True})
    assert sm.rfunc.quad


def test_stressmodel_nearest_kind_no_oseries_specified(pstore):
    with pytest.raises(ValueError, match=r"Getting nearest stress*"):
        pstore.get_stressmodel("nearest evap")  # error oseries


def test_stressmodel_nearest_no_kind_specified(pstore):
    with pytest.raises(ValueError, match=r"Could not parse stresses*"):
        pstore.get_stressmodel("nearest", oseries="oseries1")

    # also in dictionary mode
    with pytest.raises(ValueError, match=r"Could not parse stresses*"):
        pstore.get_stressmodel({"stress": ["nearest"]}, oseries="oseries1")


def test_stressmodel_nearest_kind(pstore):
    # nearest kind
    sm = pstore.get_stressmodel("nearest evap", oseries="oseries1")
    assert sm.stress[0].name == "evap1"

    # nearest kind in dict
    sm = pstore.get_stressmodel({"stress": "nearest evap"}, oseries="oseries1")
    assert sm.stress[0].name == "evap1"

    # nearest and kind separate
    sm = pstore.get_stressmodel({"stress": "nearest"}, kind="evap", oseries="oseries1")
    assert sm.stress[0].name == "evap1"

    # nearest in dict and kind separate
    sm = pstore.get_stressmodel(
        {"stress": ["nearest"]}, kind="evap", oseries="oseries1"
    )
    assert sm.stress[0].name == "evap1"


def test_recharge_model(pstore):
    # test list of stress names
    rm = pstore.get_stressmodel(["prec1", "evap1"], stressmodel="RechargeModel")
    assert rm.stress[0].name == "prec1"
    assert rm.stress[1].name == "evap1"

    # test list of nearest
    rm = pstore.get_stressmodel(
        ["nearest prec", "nearest evap"],
        stressmodel="RechargeModel",
        oseries="oseries1",
    )
    assert rm.stress[0].name == "prec1"
    assert rm.stress[1].name == "evap1"

    # test dict, no kind specified
    rm = pstore.get_stressmodel(
        {"prec": "nearest", "evap": "nearest"},
        stressmodel="RechargeModel",
        oseries="oseries1",
    )
    assert rm.stress[0].name == "prec1"
    assert rm.stress[1].name == "evap1"

    # test list, bare nearest with kind specified
    rm = pstore.get_stressmodel(
        ["nearest", "nearest"],
        kind=["prec", "evap"],
        stressmodel="RechargeModel",
        oseries="oseries1",
    )
    assert rm.stress[0].name == "prec1"
    assert rm.stress[1].name == "evap1"


def test_wellmodel(pstore):
    # test nearest <n> <kind>
    wm = pstore.get_stressmodel(
        "nearest 2 well",
        stressmodel="WellModel",
        oseries="oseries1",
    )
    assert wm.stress[0].name == "well1"
    assert wm.stress[1].name == "well2"

    # test nearest with no kind specified
    wm = pstore.get_stressmodel(
        "nearest 2",
        stressmodel="WellModel",
        oseries="oseries1",
    )
    assert wm.stress[0].name == "well1"
    assert wm.stress[1].name == "well2"

    # test nearest n, with non-existing kind specified
    with pytest.raises(ValueError, match=r"Could not find stresses*"):
        pstore.get_stressmodel(
            "nearest 2",
            kind="well2",
            stressmodel="WellModel",
            oseries="oseries1",
        )

    # test nearest n with n exceeded
    with pytest.raises(ValueError, match=r"Could not find*"):
        pstore.get_stressmodel(
            "nearest 3",
            kind="well",
            stressmodel="WellModel",
            oseries="oseries1",
        )



"""Module containing the PastaStore object for managing time series and models."""

import json
import logging
import os
import warnings
from typing import Dict, List, Literal, Optional, Tuple, Union

import numpy as np
import pandas as pd
import pastas as ps
from packaging.version import parse as parse_version
from pastas.io.pas import pastas_hook
from tqdm.auto import tqdm

from pastastore.base import BaseConnector
from pastastore.connectors import DictConnector
from pastastore.plotting import Maps, Plots
from pastastore.util import _custom_warning
from pastastore.version import PASTAS_GEQ_150, PASTAS_LEQ_022
from pastastore.yaml_interface import PastastoreYAML

FrameorSeriesUnion = Union[pd.DataFrame, pd.Series]
warnings.showwarning = _custom_warning

logger = logging.getLogger(__name__)


class PastaStore:
    """PastaStore object for managing pastas time series and models.

    Requires a Connector object to provide the interface to
    the database. Different Connectors are available, e.g.:

    - PasConnector for storing all data as .pas (JSON) files on disk (recommended)
    - ArcticDBConnector for saving data on disk using arcticdb package
    - DictConnector for storing all data in dictionaries (in-memory)

    Parameters
    ----------
    connector : Connector object
        object that provides the interface to the
        database, e.g. ArcticConnector (see pastastore.connectors)
    name : str, optional
        name of the PastaStore, by default takes the name of the Connector object
    """

    _accessors = set()

    def __init__(
        self,
        connector: Optional[BaseConnector] = None,
        name: Optional[str] = None,
    ):
        """Initialize PastaStore for managing pastas time series and models.

        Parameters
        ----------
        connector : Connector object, optional
            object that provides the connection to the database. Default is None, which
            will create a DictConnector. This default Connector does not store data on
            disk.
        name : str, optional
            name of the PastaStore, if not provided uses the Connector name
        """
        if isinstance(connector, str):
            raise DeprecationWarning(
                "PastaStore expects the connector as the first argument since v1.1!"
            )
        if connector is None:
            connector = DictConnector("pastas_db")
        self.conn = connector
        self.name = name if name is not None else self.conn.name
        self._register_connector_methods()

        # register map, plot and yaml classes
        self.maps = Maps(self)
        self.plots = Plots(self)
        self.yaml = PastastoreYAML(self)

    def _register_connector_methods(self):
        """Register connector methods (internal method)."""
        methods = [
            func
            for func in dir(self.conn)
            if callable(getattr(self.conn, func)) and not func.startswith("_")
        ]
        for meth in methods:
            setattr(self, meth, getattr(self.conn, meth))

    @property
    def oseries(self):
        """
        Returns the oseries metadata as dataframe.

        Returns
        -------
        oseries
            oseries metadata as dataframe
        """
        return self.conn.oseries

    @property
    def stresses(self):
        """
        Returns the stresses metadata as dataframe.

        Returns
        -------
        stresses
            stresses metadata as dataframe
        """
        return self.conn.stresses

    @property
    def models(self):
        """Return list of model names.

        Returns
        -------
        list
            list of model names
        """
        return self.conn.models

    @property
    def oseries_names(self):
        """Return list of oseries names.

        Returns
        -------
        list
            list of oseries names
        """
        return self.conn.oseries_names

    @property
    def stresses_names(self):
        """Return list of streses names.

        Returns
        -------
        list
            list of stresses names
        """
        return self.conn.stresses_names

    @property
    def model_names(self):
        """Return list of model names.

        Returns
        -------
        list
            list of model names
        """
        return self.conn.model_names

    @property
    def _modelnames_cache(self):
        return self.conn._modelnames_cache

    @property
    def n_oseries(self):
        """Return number of oseries.

        Returns
        -------
        int
            number of oseries
        """
        return self.conn.n_oseries

    @property
    def n_stresses(self):
        """Return number of stresses.

        Returns
        -------
        int
            number of stresses
        """
        return self.conn.n_stresses

    @property
    def n_models(self):
        """Return number of models.

        Returns
        -------
        int
            number of models
        """
        return self.conn.n_models

    @property
    def oseries_models(self):
        """Return dictionary of models per oseries.

        Returns
        -------
        dict
            dictionary containing list of models (values) for each oseries (keys).
        """
        return self.conn.oseries_models

    @property
    def oseries_with_models(self):
        """Return list of oseries for which models are contained in the database.

        Returns
        -------
        list
            list of oseries names for which models are contained in the database.
        """
        return self.conn.oseries_with_models

    def __repr__(self):
        """Representation string of the object."""
        return f"<PastaStore> {self.name}: \n - " + self.conn.__str__()

    def get_oseries_distances(
        self, names: Optional[Union[list, str]] = None
    ) -> FrameorSeriesUnion:
        """Get the distances in meters between the oseries.

        Parameters
        ----------
        names: str or list of str
            names of the oseries to calculate distances between

        Returns
        -------
        distances: pandas.DataFrame
            Pandas DataFrame with the distances between the oseries
        """
        oseries_df = self.conn.oseries
        other_df = self.conn.oseries

        names = self.conn._parse_names(names)

        xo = pd.to_numeric(oseries_df.loc[names, "x"])
        xt = pd.to_numeric(other_df.loc[:, "x"])
        yo = pd.to_numeric(oseries_df.loc[names, "y"])
        yt = pd.to_numeric(other_df.loc[:, "y"])

        xh, xi = np.meshgrid(xt, xo)
        yh, yi = np.meshgrid(yt, yo)

        distances = pd.DataFrame(
            np.sqrt((xh - xi) ** 2 + (yh - yi) ** 2),
            index=names,
            columns=other_df.index,
        )

        return distances

    def get_nearest_oseries(
        self,
        names: Optional[Union[list, str]] = None,
        n: int = 1,
        maxdist: Optional[float] = None,
    ) -> FrameorSeriesUnion:
        """Get the nearest (n) oseries.

        Parameters
        ----------
        names: str or list of str
            string or list of strings with the name(s) of the oseries
        n: int
            number of oseries to obtain
        maxdist : float, optional
            maximum distance to consider

        Returns
        -------
        oseries:
            list with the names of the oseries.
        """
        distances = self.get_oseries_distances(names)
        if maxdist is not None:
            distances = distances.where(distances <= maxdist, np.nan)

        data = pd.DataFrame(columns=np.arange(n))

        for series_name in distances.index:
            others = distances.loc[series_name].dropna().sort_values().index.tolist()
            # remove self
            others.remove(series_name)
            series = pd.DataFrame(
                index=[series_name], columns=data.columns, data=[others[:n]]
            )
            data = pd.concat([data, series], axis=0)
        return data

    def get_distances(
        self,
        oseries: Optional[Union[list, str]] = None,
        stresses: Optional[Union[list, str]] = None,
        kind: Optional[Union[str, List[str]]] = None,
    ) -> FrameorSeriesUnion:
        """Get the distances in meters between the oseries and stresses.

        Parameters
        ----------
        oseries: str or list of str
            name(s) of the oseries
        stresses: str or list of str
            name(s) of the stresses
        kind: str, list of str
            string or list of strings representing which kind(s) of
            stresses to consider

        Returns
        -------
        distances: pandas.DataFrame
            Pandas DataFrame with the distances between the oseries (index)
            and the stresses (columns).
        """
        oseries_df = self.conn.oseries
        stresses_df = self.conn.stresses

        oseries = self.conn._parse_names(oseries)

        if stresses is None and kind is None:
            stresses = stresses_df.index
        elif stresses is not None and kind is not None:
            if isinstance(kind, str):
                kind = [kind]
            mask = stresses_df.kind.isin(kind)
            stresses = stresses_df.loc[stresses].loc[mask].index
        elif stresses is None:
            if isinstance(kind, str):
                kind = [kind]
            stresses = stresses_df.loc[stresses_df.kind.isin(kind)].index

        xo = pd.to_numeric(oseries_df.loc[oseries, "x"])
        xt = pd.to_numeric(stresses_df.loc[stresses, "x"])
        yo = pd.to_numeric(oseries_df.loc[oseries, "y"])
        yt = pd.to_numeric(stresses_df.loc[stresses, "y"])

        xh, xi = np.meshgrid(xt, xo)
        yh, yi = np.meshgrid(yt, yo)

        distances = pd.DataFrame(
            np.sqrt((xh - xi) ** 2 + (yh - yi) ** 2),
            index=oseries,
            columns=stresses,
        )

        return distances

    def get_nearest_stresses(
        self,
        oseries: Optional[Union[list, str]] = None,
        stresses: Optional[Union[list, str]] = None,
        kind: Optional[Union[list, str]] = None,
        n: int = 1,
        maxdist: Optional[float] = None,
    ) -> FrameorSeriesUnion:
        """Get the nearest (n) stresses of a specific kind.

        Parameters
        ----------
        oseries: str
            string with the name of the oseries
        stresses: str or list of str
            string with the name of the stresses
        kind: str, list of str, optional
            string or list of str with the name of the kind(s)
            of stresses to consider
        n: int
            number of stresses to obtain
        maxdist : float, optional
            maximum distance to consider

        Returns
        -------
        stresses:
            list with the names of the stresses.
        """
        distances = self.get_distances(oseries, stresses, kind)
        if maxdist is not None:
            distances = distances.where(distances <= maxdist, np.nan)

        data = pd.DataFrame(columns=np.arange(n))

        for series in distances.index:
            series = pd.DataFrame(
                [distances.loc[series].dropna().sort_values().index[:n]]
            )
            data = pd.concat([data, series], axis=0)
        return data

    def get_signatures(
        self,
        signatures=None,
        names=None,
        libname="oseries",
        progressbar=False,
        ignore_errors=False,
    ):
        """Get groundwater signatures.

        NaN-values are returned when the signature cannot be computed.

        Parameters
        ----------
        signatures : list of str, optional
            list of groundwater signatures to compute, if None all groundwater
            signatures in ps.stats.signatures.__all__ are used, by default None
        names : str, list of str, or None, optional
            names of the time series, by default None which
            uses all the time series in the library
        libname : str
            name of the library containing the time series
            ('oseries' or 'stresses'), by default "oseries"
        progressbar : bool, optional
            show progressbar, by default False
        ignore_errors : bool, optional
            ignore errors when True, i.e. when non-existent timeseries is
            encountered in names, by default False

        Returns
        -------
        signatures_df : pandas.DataFrame
            DataFrame containing the signatures (columns) per time series (rows)
        """
        names = self.conn._parse_names(names, libname=libname)

        if signatures is None:
            signatures = ps.stats.signatures.__all__.copy()

        # create dataframe for results
        signatures_df = pd.DataFrame(index=names, columns=signatures, data=np.nan)

        # loop through oseries names
        desc = "Get groundwater signatures"
        for name in tqdm(names, desc=desc) if progressbar else names:
            try:
                if libname == "oseries":
                    s = self.conn.get_oseries(name)
                else:
                    s = self.conn.get_stresses(name)
            except Exception as e:
                if ignore_errors:
                    signatures_df.loc[name, :] = np.nan
                    continue
                else:
                    raise e

            try:
                i_signatures = ps.stats.signatures.summary(s.squeeze(), signatures)
            except Exception as e:
                if ignore_errors:
                    i_signatures = []
                    for signature in signatures:
                        try:
                            sign_val = getattr(ps.stats.signatures, signature)(
                                s.squeeze()
                            )
                        except Exception as _:
                            sign_val = np.nan
                        i_signatures.append(sign_val)
                else:
                    raise e
            signatures_df.loc[name, signatures] = i_signatures.squeeze()

        return signatures_df

    def get_tmin_tmax(
        self,
        libname: Literal["oseries", "stresses", "models"],
        names: Union[str, List[str], None] = None,
        progressbar: bool = False,
    ):
        """Get tmin and tmax for time series.

        Parameters
        ----------
        libname : str
            name of the library containing the time series
            ('oseries' or 'stresses')
        names : str, list of str, or None, optional
            names of the time series, by default None which
            uses all the time series in the library
        progressbar : bool, optional
            show progressbar, by default False

        Returns
        -------
        tmintmax : pd.dataframe
            Dataframe containing tmin and tmax per time series
        """
        names = self.conn._parse_names(names, libname=libname)
        tmintmax = pd.DataFrame(
            index=names, columns=["tmin", "tmax"], dtype="datetime64[ns]"
        )
        desc = f"Get tmin/tmax {libname}"
        for n in tqdm(names, desc=desc) if progressbar else names:
            if libname == "models":
                mld = self.conn.get_models(
                    n,
                    return_dict=True,
                )
                tmintmax.loc[n, "tmin"] = mld["settings"]["tmin"]
                tmintmax.loc[n, "tmax"] = mld["settings"]["tmax"]
            else:
                s = (
                    self.conn.get_oseries(n)
                    if libname == "oseries"
                    else self.conn.get_stresses(n)
                )
                tmintmax.loc[n, "tmin"] = s.first_valid_index()
                tmintmax.loc[n, "tmax"] = s.last_valid_index()

        return tmintmax

    def get_extent(self, libname, names=None, buffer=0.0):
        """Get extent [xmin, xmax, ymin, ymax] from library.

        Parameters
        ----------
        libname : str
            name of the library containing the time series
            ('oseries', 'stresses', 'models')
        names : str, list of str, or None, optional
            list of names to include for computing the extent
        buffer : float, optional
            add this distance to the extent, by default 0.0

        Returns
        -------
        extent : list
            extent [xmin, xmax, ymin, ymax]
        """
        names = self.conn._parse_names(names, libname=libname)
        if libname in ["oseries", "stresses"]:
            df = getattr(self, libname)
        elif libname == "models":
            df = self.oseries
        else:
            raise ValueError(f"Cannot get extent for library '{libname}'.")

        extent = [
            df.loc[names, "x"].min() - buffer,
            df.loc[names, "x"].max() + buffer,
            df.loc[names, "y"].min() - buffer,
            df.loc[names, "y"].max() + buffer,
        ]
        return extent

    def get_parameters(
        self,
        parameters: Optional[List[str]] = None,
        modelnames: Optional[List[str]] = None,
        param_value: Optional[str] = "optimal",
        progressbar: Optional[bool] = False,
        ignore_errors: Optional[bool] = False,
    ) -> FrameorSeriesUnion:
        """Get model parameters.

        NaN-values are returned when the parameters are not present in the model or the
        model is not optimized.

        Parameters
        ----------
        parameters : list of str, optional
            names of the parameters, by default None which uses all
            parameters from each model
        modelnames : str or list of str, optional
            name(s) of model(s), by default None in which case all models
            are used
        param_value : str, optional
            which column to use from the model parameters dataframe, by
            default "optimal" which retrieves the optimized parameters.
        progressbar : bool, optional
            show progressbar, default is False
        ignore_errors : bool, optional
            ignore errors when True, i.e. when non-existent model is
            encountered in modelnames, by default False

        Returns
        -------
        p : pandas.DataFrame
            DataFrame containing the parameters (columns) per model (rows)
        """
        modelnames = self.conn._parse_names(modelnames, libname="models")

        # create dataframe for results
        p = pd.DataFrame(index=modelnames, columns=parameters)

        # loop through model names and store results
        desc = "Get model parameters"
        for mlname in tqdm(modelnames, desc=desc) if progressbar else modelnames:
            try:
                mldict = self.get_models(mlname, return_dict=True, progressbar=False)
            except Exception as e:
                if ignore_errors:
                    p.loc[mlname, :] = np.nan
                    continue
                else:
                    raise e
            if parameters is None:
                pindex = mldict["parameters"].index
            else:
                pindex = parameters

            for c in pindex:
                p.loc[mlname, c] = mldict["parameters"].loc[c, param_value]

        p = p.squeeze()
        return p.astype(float)

    def get_statistics(
        self,
        statistics: Union[str, List[str]],
        modelnames: Optional[List[str]] = None,
        progressbar: Optional[bool] = False,
        ignore_errors: Optional[bool] = False,
        **kwargs,
    ) -> FrameorSeriesUnion:
        """Get model statistics.

        Parameters
        ----------
        statistics : str or list of str
            statistic or list of statistics to calculate, e.g. ["evp", "rsq", "rmse"],
            for a full list see `pastas.modelstats.Statistics.ops`.
        modelnames : list of str, optional
            modelnames to calculates statistics for, by default None, which
            uses all models in the store
        progressbar : bool, optional
            show progressbar, by default False
        ignore_errors : bool, optional
            ignore errors when True, i.e. when trying to calculate statistics
            for non-existent model in modelnames, default is False
        **kwargs
            any arguments that can be passed to the methods for calculating
            statistics

        Returns
        -------
        s : pandas.DataFrame
        """
        modelnames = self.conn._parse_names(modelnames, libname="models")

        # if statistics is str
        if isinstance(statistics, str):
            statistics = [statistics]

        # create dataframe for results
        s = pd.DataFrame(index=modelnames, columns=statistics, data=np.nan)

        # loop through model names
        desc = "Get model statistics"
        for mlname in tqdm(modelnames, desc=desc) if progressbar else modelnames:
            try:
                ml = self.get_models(mlname, progressbar=False)
            except Exception as e:
                if ignore_errors:
                    continue
                else:
                    raise e
            for stat in statistics:
                value = ml.stats.__getattribute__(stat)(**kwargs)
                s.loc[mlname, stat] = value

        s = s.squeeze()
        return s.astype(float)

    def create_model(
        self,
        name: str,
        modelname: Optional[str] = None,
        add_recharge: bool = True,
        add_ar_noisemodel: bool = False,
        recharge_name: str = "recharge",
    ) -> ps.Model:
        """Create a pastas Model.

        Parameters
        ----------
        name : str
            name of the oseries to create a model for
        modelname : str, optional
            name of the model, default is None, which uses oseries name
        add_recharge : bool, optional
            add recharge to the model by looking for the closest
            precipitation and evaporation time series in the stresses
            library, by default True
        add_ar_noisemodel : bool, optional
            add AR(1) noise model to the model, by default False
        recharge_name : str
            name of the RechargeModel

        Returns
        -------
        pastas.Model
            model for the oseries

        Raises
        ------
        KeyError
            if data is stored as dataframe and no column is provided
        ValueError
            if time series is empty
        """
        # get oseries metadata
        meta = self.conn.get_metadata("oseries", name, as_frame=False)
        ts = self.conn.get_oseries(name)

        # convert to time series and create model
        if not ts.dropna().empty:
            if modelname is None:
                modelname = name
            ml = ps.Model(ts, name=modelname, metadata=meta)
            if add_recharge:
                self.add_recharge(ml, recharge_name=recharge_name)
            if add_ar_noisemodel and PASTAS_GEQ_150:
                ml.add_noisemodel(ps.ArNoiseModel())
            return ml
        else:
            raise ValueError("Empty time series!")

    def create_models_bulk(
        self,
        oseries: Optional[Union[list, str]] = None,
        add_recharge: bool = True,
        solve: bool = False,
        store_models: bool = True,
        ignore_errors: bool = False,
        progressbar: bool = True,
        **kwargs,
    ) -> Union[Tuple[dict, dict], dict]:
        """Bulk creation of pastas models.

        Parameters
        ----------
        oseries : list of str, optional
            names of oseries to create models for, by default None,
            which creates models for all oseries
        add_recharge : bool, optional
            add recharge to the models based on closest
            precipitation and evaporation time series, by default True
        solve : bool, optional
            solve the model, by default False
        store_models : bool, optional
            if False, return a list of models, by default True, which will
            store the models in the database.
        ignore_errors : bool, optional
            ignore errors while creating models, by default False
        progressbar : bool, optional
            show progressbar, by default True

        Returns
        -------
        models : dict, if return_models is True
            dictionary of models
        errors : list, always returned
            list of model names that could not be created
        """
        if oseries is None:
            oseries = self.conn.oseries.index
        elif isinstance(oseries, str):
            oseries = [oseries]

        models = {}
        errors = {}
        desc = "Bulk creation models"
        for o in tqdm(oseries, desc=desc) if progressbar else oseries:
            try:
                iml = self.create_model(o, add_recharge=add_recharge)
            except Exception as e:
                if ignore_errors:
                    errors[o] = e
                    continue
                else:
                    raise e
            if solve:
                iml.solve(**kwargs)
            if store_models:
                self.conn.add_model(iml, overwrite=True)
            else:
                models[o] = iml
        if len(errors) > 0:
            print("Warning! Errors occurred while creating models!")
        if store_models:
            return errors
        else:
            return models, errors

    def add_recharge(
        self,
        ml: ps.Model,
        rfunc=None,
        recharge=None,
        recharge_name: str = "recharge",
    ) -> None:
        """Add recharge to a pastas model.

        Uses closest precipitation and evaporation time series in database.
        These are assumed to be labeled with kind = 'prec' or 'evap'.

        Parameters
        ----------
        ml : pastas.Model
            pastas.Model object
        rfunc : pastas.rfunc, optional
            response function to use for recharge in model, by default None
            which uses ps.Exponential() (for different response functions, see pastas
            documentation)
        recharge : ps.RechargeModel
            recharge model to use, default is ps.rch.Linear()
        recharge_name : str
            name of the RechargeModel
        """
        if recharge is None:
            recharge = ps.rch.Linear()
        if rfunc is None:
            rfunc = ps.Exponential

        self.add_stressmodel(
            ml,
            stresses={"prec": "nearest", "evap": "nearest"},
            rfunc=rfunc,
            stressmodel=ps.RechargeModel,
            stressmodel_name=recharge_name,
            recharge=recharge,
        )

    def _parse_stresses(
        self,
        stresses: Union[str, List[str], Dict[str, str]],
        kind: Optional[str],
        stressmodel,
        oseries: Optional[str] = None,
    ):
        # parse stresses for RechargeModel, allow list of len 2 or 3 and
        # set correct kwarg names
        if stressmodel._name == "RechargeModel":
            if isinstance(stresses, list):
                if len(stresses) == 2:
                    stresses = {
                        "prec": stresses[0],
                        "evap": stresses[1],
                    }
                elif len(stresses) == 3:
                    stresses = {
                        "prec": stresses[0],
                        "evap": stresses[1],
                        "temp": stresses[2],
                    }
                else:
                    raise ValueError(
                        "RechargeModel requires 2 or 3 stress names, "
                        f"got: {len(stresses)}!"
                    )
        # if stresses is list, create dictionary normally
        elif isinstance(stresses, list):
            stresses = {"stress": stresses}
        # if stresses is str, make it a list of len 1
        elif isinstance(stresses, str):
            stresses = {"stress": [stresses]}

        # check if stresses is a dictionary, else raise TypeError
        if not isinstance(stresses, dict):
            raise TypeError("stresses must be a list, string or dictionary!")

        # if no kind specified, set to well for WellModel
        if stressmodel._name == "WellModel":
            if kind is None:
                kind = "well"

        # store a copy of the user input for kind
        if isinstance(kind, list):
            _kind = kind.copy()
        else:
            _kind = kind

        # create empty list for gathering metadata
        metadata = []
        # loop over stresses keys/values
        for i, (k, v) in enumerate(stresses.items()):
            # if entry in dictionary is str, make it list of len 1
            if isinstance(v, str):
                v = [v]
            # parse value
            if isinstance(v, list):
                for item in v:
                    names = []  # empty list for names
                    # parse nearest
                    if item.startswith("nearest"):
                        # check oseries defined if nearest option is used
                        if not oseries:
                            raise ValueError(
                                "Getting nearest stress(es) requires oseries name!"
                            )
                        try:
                            if len(item.split()) == 3:  # nearest <n> <kind>
                                n = int(item.split()[1])
                                kind = item.split()[2]
                            elif len(item.split()) == 2:  # nearest <n> | <kind>
                                try:
                                    n = int(item.split()[1])  # try converting to <n>
                                except ValueError:
                                    n = 1
                                    kind = item.split()[1]  # interpret as <kind>
                            else:  # nearest
                                n = 1
                                # if RechargeModel, we can infer kind
                                if (
                                    _kind is None
                                    and stressmodel._name == "RechargeModel"
                                ):
                                    kind = k
                                elif _kind is None:  # catch no kind with bare nearest
                                    raise ValueError(
                                        "Bare 'nearest' found but no kind specified."
                                    )
                                elif isinstance(_kind, list):
                                    kind = _kind[i]  # if multiple kind, select i-th
                        except Exception as e:
                            # raise if nearest parsing failed
                            raise ValueError(
                                f"Could not parse stresses: '{item}'! "
                                "When using option 'nearest', use 'nearest' and specify"
                                " kind, or 'nearest <kind>' or 'nearest <n> <kind>'!"
                            ) from e
                        # check if kind exists at all
                        if kind not in self.stresses.kind.values:
                            raise ValueError(
                                f"Could not find stresses with kind='{kind}'!"
                            )
                        # get stress names of <n> nearest <kind> stresses
                        inames = self.get_nearest_stresses(
                            oseries, kind=kind, n=n
                        ).iloc[0]
                        # check if any NaNs in result
                        if inames.isna().any():
                            nkind = (self.stresses.kind == kind).sum()
                            raise ValueError(
                                f"Could not find {n} nearest stress(es) for '{kind}'! "
                                f"There are only {nkind} '{kind}' stresses."
                            )
                        # append names
                        names += inames.tolist()
                    else:
                        # assume name is name of stress
                        names.append(item)
                # get stresses and metadata
                stress_series, imeta = self.get_stresses(
                    names, return_metadata=True, squeeze=True
                )
                # replace stress name(s) with time series
                if len(names) > 1:
                    stresses[k] = list(stress_series.values())
                else:
                    stresses[k] = stress_series
                # gather metadata
                if isinstance(imeta, list):
                    metadata += imeta
                else:
                    metadata.append(imeta)

        return stresses, metadata

    def get_stressmodel(
        self,
        stresses: Union[str, List[str], Dict[str, str]],
        stressmodel=ps.StressModel,
        stressmodel_name: Optional[str] = None,
        rfunc=ps.Exponential,
        rfunc_kwargs: Optional[dict] = None,
        kind: Optional[Union[List[str], str]] = None,
        oseries: Optional[str] = None,
        **kwargs,
    ):
        """Get a Pastas stressmodel from stresses time series in Pastastore.

        Supports "nearest" selection. Any stress name can be replaced by
        "nearest [<n>] <kind>" where <n> is optional and represents the number of
        nearest stresses and <kind> and represents the kind of stress to
        consider. <kind> can also be specified directly with the `kind` kwarg.

        Note: the 'nearest' option requires the oseries name to be provided.
        Additionally, 'x' and 'y' metadata must be stored for oseries and stresses.

        Parameters
        ----------
        stresses : str, list of str, or dict
            name(s) of the time series to use for the stressmodel, or dictionary
            with key(s) and value(s) as time series name(s). Options include:
               - name of stress: `"prec_stn"`
               - list of stress names: `["prec_stn", "evap_stn"]`
               - dict for RechargeModel: `{"prec": "prec_stn", "evap": "evap_stn"}`
               - dict for StressModel: `{"stress": "well1"}`
               - nearest, specifying kind: `"nearest well"`
               - nearest specifying number and kind: `"nearest 2 well"`
        stressmodel : str or class
            stressmodel class to use, by default ps.StressModel
        stressmodel_name : str, optional
            name of the stressmodel, by default None, which uses the stress name,
            if there is 1 stress otherwise the name of the stressmodel type. For
            RechargeModels, the name defaults to 'recharge'.
        rfunc : str or class
            response function class to use, by default ps.Exponential
        rfunc_kwargs : dict, optional
            keyword arguments to pass to the response function, by default None
        kind : str or list of str, optional
            specify kind of stress(es) to use, by default None, useful in combination
            with 'nearest' option for defining stresses
        oseries : str, optional
            name of the oseries to use for the stressmodel, by default None, used when
            'nearest' option is used for defining stresses.
        **kwargs
            additional keyword arguments to pass to the stressmodel

        Returns
        -------
        stressmodel : pastas.StressModel
            pastas StressModel that can be added to pastas Model.
        """
        # get stressmodel class, if str was provided
        if isinstance(stressmodel, str):
            stressmodel = getattr(ps, stressmodel)

        # parse stresses names to get time series and metadata
        stresses, metadata = self._parse_stresses(
            stresses=stresses, stressmodel=stressmodel, kind=kind, oseries=oseries
        )

        # get stressmodel name if not provided
        if stressmodel_name is None:
            if stressmodel._name == "RechargeModel":
                stressmodel_name = "recharge"
            elif len(metadata) == 1:
                stressmodel_name = stresses["stress"].squeeze().name
            else:
                stressmodel_name = stressmodel._name

        # check if metadata is list of len 1 and unpack
        if isinstance(metadata, list) and len(metadata) == 1:
            metadata = metadata[0]

        # get stressmodel time series settings
        if kind and "settings" not in kwargs:
            # try using kind to get predefined settings options
            if isinstance(kind, str):
                kwargs["settings"] = ps.rcParams["timeseries"].get(kind, None)
            else:
                kwargs["settings"] = [
                    ps.rcParams["timeseries"].get(ikind, None) for ikind in kind
                ]
        elif kind is None and "settings" not in kwargs:
            # try using kind stored in metadata to get predefined settings options
            if isinstance(metadata, list):
                kwargs["settings"] = [
                    ps.rcParams["timeseries"].get(imeta.get("kind", None), None)
                    for imeta in metadata
                ]
            elif isinstance(metadata, dict):
                kwargs["settings"] = ps.rcParams["timeseries"].get(
                    metadata.get("kind", None), None
                )

        # get rfunc class if str was provided
        if isinstance(rfunc, str):
            rfunc = getattr(ps, rfunc)

        # create empty rfunc_kwargs if not provided
        if rfunc_kwargs is None:
            rfunc_kwargs = {}

        # special for WellModels
        if stressmodel._name == "WellModel":
            names = [s.squeeze().name for s in stresses["stress"]]
            # check oseries is provided
            if oseries is None:
                raise ValueError("WellModel requires 'oseries' to compute distances!")
            # compute distances and add to kwargs
            distances = (
                self.get_distances(oseries=oseries, stresses=names).T.squeeze().values
            )
            kwargs["distances"] = distances
            # set settings to well
            if "settings" not in kwargs:
                kwargs["settings"] = "well"
            # override rfunc and set to HantushWellModel
            rfunc = ps.HantushWellModel

        # do not add metadata for pastas 0.22 and WellModel
        if not PASTAS_LEQ_022 and (stressmodel._name != "WellModel"):
            kwargs["metadata"] = metadata

        return stressmodel(
            **stresses,
            rfunc=rfunc(**rfunc_kwargs),
            name=stressmodel_name,
            **kwargs,
        )

    def add_stressmodel(
        self,
        ml: Union[ps.Model, str],
        stresses: Union[str, List[str], Dict[str, str]],
        stressmodel=ps.StressModel,
        stressmodel_name: Optional[str] = None,
        rfunc=ps.Exponential,
        rfunc_kwargs: Optional[dict] = None,
        kind: Optional[Union[List[str], str]] = None,
        **kwargs,
    ):
        """Add a pastas StressModel from stresses time series in Pastastore.

        Supports "nearest" selection. Any stress name can be replaced by
        "nearest [<n>] <kind>" where <n> is optional and represents the number of
        nearest stresses and <kind> and represents the kind of stress to
        consider. <kind> can also be specified directly with the `kind` kwarg.

        Note: the 'nearest' option requires the oseries name to be provided.
        Additionally, 'x' and 'y' metadata must be stored for oseries and stresses.

        Parameters
        ----------
        ml : pastas.Model or str
            pastas.Model object to add StressModel to, if passed as string,
            model is loaded from store, the stressmodel is added and then written
            back to the store.
        stresses : str, list of str, or dict
            name(s) of the time series to use for the stressmodel, or dictionary
            with key(s) and value(s) as time series name(s). Options include:
               - name of stress: `"prec_stn"`
               - list of stress names: `["prec_stn", "evap_stn"]`
               - dict for RechargeModel: `{"prec": "prec_stn", "evap": "evap_stn"}`
               - dict for StressModel: `{"stress": "well1"}`
               - nearest, specifying kind: `"nearest well"`
               - nearest specifying number and kind: `"nearest 2 well"`
        stressmodel : str or class
            stressmodel class to use, by default ps.StressModel
        stressmodel_name : str, optional
            name of the stressmodel, by default None, which uses the stress name,
            if there is 1 stress otherwise the name of the stressmodel type. For
            RechargeModels, the name defaults to 'recharge'.
        rfunc : str or class
            response function class to use, by default ps.Exponential
        rfunc_kwargs : dict, optional
            keyword arguments to pass to the response function, by default None
        kind : str or list of str, optional
            specify kind of stress(es) to use, by default None, useful in combination
            with 'nearest' option for defining stresses
        **kwargs
            additional keyword arguments to pass to the stressmodel
        """
        sm = self.get_stressmodel(
            stresses=stresses,
            stressmodel=stressmodel,
            stressmodel_name=stressmodel_name,
            rfunc=rfunc,
            rfunc_kwargs=rfunc_kwargs,
            kind=kind,
            oseries=ml if isinstance(ml, str) else ml.oseries.name,
            **kwargs,
        )
        if isinstance(ml, str):
            ml = self.get_model(ml)
            ml.add_stressmodel(sm)
            self.conn.add_model(ml, overwrite=True)
            logger.info(
                f"Stressmodel '{sm.name}' added to model '{ml.name}' "
                "and stored in database."
            )
        else:
            ml.add_stressmodel(sm)

    def solve_models(
        self,
        mls: Optional[Union[ps.Model, list, str]] = None,
        report: bool = False,
        ignore_solve_errors: bool = False,
        store_result: bool = True,
        progressbar: bool = True,
        **kwargs,
    ) -> None:
        """Solves the models in the store.

        Parameters
        ----------
        mls : list of str, optional
            list of model names, if None all models in the pastastore
            are solved.
        report : boolean, optional
            determines if a report is printed when the model is solved,
            default is False
        ignore_solve_errors : boolean, optional
            if True, errors emerging from the solve method are ignored,
            default is False which will raise an exception when a model
            cannot be optimized
        store_result : bool, optional
            if True save optimized models, default is True
        progressbar : bool, optional
            show progressbar, default is True
        **kwargs :
            arguments are passed to the solve method.
        """
        if mls is None:
            mls = self.conn.model_names
        elif isinstance(mls, ps.Model):
            mls = [mls.name]

        desc = "Solving models"
        for ml_name in tqdm(mls, desc=desc) if progressbar else mls:
            ml = self.conn.get_models(ml_name)

            m_kwargs = {}
            for key, value in kwargs.items():
                if isinstance(value, pd.Series):
                    m_kwargs[key] = value.loc[ml_name]
                else:
                    m_kwargs[key] = value
            # Convert timestamps
            for tstamp in ["tmin", "tmax"]:
                if tstamp in m_kwargs:
                    m_kwargs[tstamp] = pd.Timestamp(m_kwargs[tstamp])

            try:
                ml.solve(report=report, **m_kwargs)
                if store_result:
                    self.conn.add_model(ml, overwrite=True)
            except Exception as e:
                if ignore_solve_errors:
                    warning = "solve error ignored for -> {}".format(ml.name)
                    ps.logger.warning(warning)
                else:
                    raise e

    def model_results(
        self,
        mls: Optional[Union[ps.Model, list, str]] = None,
        progressbar: bool = True,
    ):  # pragma: no cover
        """Get pastas model results.

        Parameters
        ----------
        mls : list of str, optional
            list of model names, by default None which means results for
            all models will be calculated
        progressbar : bool, optional
            show progressbar, by default True

        Returns
        -------
        results : pd.DataFrame
            dataframe containing parameters and other statistics
            for each model

        Raises
        ------
        ModuleNotFoundError
            if the art_tools module is not available
        """
        try:
            from art_tools import pastas_get_model_results
        except Exception as e:
            raise ModuleNotFoundError("You need 'art_tools' to use this method!") from e

        if mls is None:
            mls = self.conn.models
        elif isinstance(mls, ps.Model):
            mls = [mls.name]

        results_list = []
        desc = "Get model results"
        for mlname in tqdm(mls, desc=desc) if progressbar else mls:
            try:
                iml = self.conn.get_models(mlname)
            except Exception as e:
                print("{1}: '{0}' could not be parsed!".format(mlname, e))
                continue
            iresults = pastas_get_model_results(
                iml, par_selection="all", stats=("evp",), stderrors=True
            )
            results_list.append(iresults)

        return pd.concat(results_list, axis=1).transpose()

    def to_zip(self, fname: str, overwrite=False, progressbar: bool = True):
        """Write data to zipfile.

        Parameters
        ----------
        fname : str
            name of zipfile
        overwrite : bool, optional
            if True, overwrite existing file
        progressbar : bool, optional
            show progressbar, by default True
        """
        from zipfile import ZIP_DEFLATED, ZipFile

        if os.path.exists(fname) and not overwrite:
            raise FileExistsError(
                "File already exists! " "Use 'overwrite=True' to " "force writing file."
            )
        elif os.path.exists(fname):
            warnings.warn(f"Overwriting file '{os.path.basename(fname)}'", stacklevel=1)

        with ZipFile(fname, "w", compression=ZIP_DEFLATED) as archive:
            # oseries
            self.conn._series_to_archive(archive, "oseries", progressbar=progressbar)
            # stresses
            self.conn._series_to_archive(archive, "stresses", progressbar=progressbar)
            # models
            self.conn._models_to_archive(archive, progressbar=progressbar)

    def export_model_series_to_csv(
        self,
        names: Optional[Union[list, str]] = None,
        exportdir: str = ".",
        exportmeta: bool = True,
    ):  # pragma: no cover
        """Export model time series to csv files.

        Parameters
        ----------
        names : Optional[Union[list, str]], optional
            names of models to export, by default None, which uses retrieves
            all models from database
        exportdir : str, optional
            directory to export csv files to, default is current directory
        exportmeta : bool, optional
            export metadata for all time series as csv file, default is True
        """
        names = self.conn._parse_names(names, libname="models")
        for name in names:
            mldict = self.get_models(name, return_dict=True)

            oname = mldict["oseries"]["name"]
            o = self.get_oseries(oname)
            o.to_csv(os.path.join(exportdir, f"{oname}.csv"))

            if exportmeta:
                metalist = [self.get_metadata("oseries", oname)]

            for sm in mldict["stressmodels"]:
                if mldict["stressmodels"][sm]["stressmodel"] == "RechargeModel":
                    for istress in ["prec", "evap"]:
                        istress = mldict["stressmodels"][sm][istress]
                        stress_name = istress["name"]
                        ts = self.get_stresses(stress_name)
                        ts.to_csv(os.path.join(exportdir, f"{stress_name}.csv"))
                        if exportmeta:
                            tsmeta = self.get_metadata("stresses", stress_name)
                            metalist.append(tsmeta)
                else:
                    for istress in mldict["stressmodels"][sm]["stress"]:
                        stress_name = istress["name"]
                        ts = self.get_stresses(stress_name)
                        ts.to_csv(os.path.join(exportdir, f"{stress_name}.csv"))
                        if exportmeta:
                            tsmeta = self.get_metadata("stresses", stress_name)
                            metalist.append(tsmeta)

            if exportmeta:
                pd.concat(metalist, axis=0).to_csv(
                    os.path.join(exportdir, f"metadata_{name}.csv")
                )

    @classmethod
    def from_zip(
        cls,
        fname: str,
        conn: Optional[BaseConnector] = None,
        storename: Optional[str] = None,
        progressbar: bool = True,
    ):
        """Load PastaStore from zipfile.

        Parameters
        ----------
        fname : str
            pathname of zipfile
        conn : Connector object, optional
            connector for storing loaded data, default is None which creates a
            DictConnector. This Connector does not store data on disk.
        storename : str, optional
            name of the PastaStore, by default None, which
            defaults to the name of the Connector.
        progressbar : bool, optional
            show progressbar, by default True

        Returns
        -------
        pastastore.PastaStore
            return PastaStore containing data from zipfile
        """
        from zipfile import ZipFile

        if conn is None:
            conn = DictConnector("pastas_db")

        with ZipFile(fname, "r") as archive:
            namelist = [
                fi for fi in archive.namelist() if not fi.endswith("_meta.json")
            ]
            for f in tqdm(namelist, desc="Reading zip") if progressbar else namelist:
                libname, fjson = os.path.split(f)
                if libname in ["stresses", "oseries"]:
                    s = pd.read_json(archive.open(f), dtype=float, orient="columns")
                    if not isinstance(s.index, pd.DatetimeIndex):
                        s.index = pd.to_datetime(s.index, unit="ms")
                    s = s.sort_index()
                    meta = json.load(archive.open(f.replace(".json", "_meta.json")))
                    conn._add_series(libname, s, fjson.split(".")[0], metadata=meta)
                elif libname in ["models"]:
                    ml = json.load(archive.open(f), object_hook=pastas_hook)
                    conn.add_model(ml)
        if storename is None:
            storename = conn.name
        return cls(conn, storename)

    def search(
        self,
        libname: str,
        s: Optional[Union[list, str]] = None,
        case_sensitive: bool = True,
        sort=True,
    ):
        """Search for names of time series or models starting with `s`.

        Parameters
        ----------
        libname : str
            name of the library to search in
        s : str, lst
            find names with part of this string or strings in list
        case_sensitive : bool, optional
            whether search should be case sensitive, by default True
        sort : bool, optional
            sort list of names

        Returns
        -------
        matches : list
            list of names that match search result
        """
        if libname == "models":
            lib_names = self.model_names
        elif libname == "stresses":
            lib_names = self.stresses_names
        elif libname == "oseries":
            lib_names = self.oseries_names
        else:
            raise ValueError("Provide valid libname: 'models', 'stresses' or 'oseries'")

        if isinstance(s, str):
            if case_sensitive:
                matches = [n for n in lib_names if s in n]
            else:
                matches = [n for n in lib_names if s.lower() in n.lower()]
        if isinstance(s, list):
            m = np.array([])
            for sub in s:
                if case_sensitive:
                    m = np.append(m, [n for n in lib_names if sub in n])
                else:
                    m = np.append(m, [n for n in lib_names if sub.lower() in n.lower()])
            matches = list(np.unique(m))
        if sort:
            matches.sort()
        return matches

    def get_model_timeseries_names(
        self,
        modelnames: Optional[Union[list, str]] = None,
        dropna: bool = True,
        progressbar: bool = True,
    ) -> FrameorSeriesUnion:
        """Get time series names contained in model.

        Parameters
        ----------
        modelnames : Optional[Union[list, str]], optional
            list or name of models to get time series names for,
            by default None which will use all modelnames
        dropna : bool, optional
            drop stresses from table if stress is not included in any
            model, by default True
        progressbar : bool, optional
            show progressbar, by default True

        Returns
        -------
        structure : pandas.DataFrame
            returns DataFrame with oseries name per model, and a flag
            indicating whether a stress is contained within a time series
            model.
        """
        model_names = self.conn._parse_names(modelnames, libname="models")
        structure = pd.DataFrame(
            index=model_names, columns=["oseries"] + self.stresses_names
        )

        for mlnam in (
            tqdm(model_names, desc="Get model time series names")
            if progressbar
            else model_names
        ):
            iml = self.get_models(mlnam, return_dict=True)

            PASFILE_LEQ_022 = parse_version(
                iml["file_info"]["pastas_version"]
            ) <= parse_version("0.22.0")

            # oseries
            structure.loc[mlnam, "oseries"] = iml["oseries"]["name"]

            for sm in iml["stressmodels"].values():
                class_key = "stressmodel" if PASFILE_LEQ_022 else "class"
                if sm[class_key] == "RechargeModel":
                    pnam = sm["prec"]["name"]
                    enam = sm["evap"]["name"]
                    structure.loc[mlnam, pnam] = 1
                    structure.loc[mlnam, enam] = 1
                elif "stress" in sm:
                    smstress = sm["stress"]
                    if isinstance(smstress, dict):
                        smstress = [smstress]
                    for s in smstress:
                        structure.loc[mlnam, s["name"]] = 1
        if dropna:
            return structure.dropna(how="all", axis=1)
        else:
            return structure

    def apply(self, libname, func, names=None, progressbar=True):
        """Apply function to items in library.

        Supported libraries are oseries, stresses, and models.

        Parameters
        ----------
        libname : str
            library name, supports "oseries", "stresses" and "models"
        func : callable
            function that accepts items from one of the supported libraries as input
        names : str, list of str, optional
            apply function to these names, by default None which loops over all stored
            items in library
        progressbar : bool, optional
            show progressbar, by default True

        Returns
        -------
        dict
            dict of results of func, with names as keys and results as values
        """
        names = self.conn._parse_names(names, libname)
        result = {}
        if libname not in ("oseries", "stresses", "models"):
            raise ValueError(
                "'libname' must be one of ['oseries', 'stresses', 'models']!"
            )
        getter = getattr(self.conn, f"get_{libname}")
        for n in (
            tqdm(names, desc=f"Applying {func.__name__}") if progressbar else names
        ):
            result[n] = func(getter(n))
        return result

    def within(self, extent, names=None, libname="oseries"):
        """Get names of items within extent.

        Parameters
        ----------
        extent : list
            list with [xmin, xmax, ymin, ymax]
        names : str, list of str, optional
            list of names to include, by default None
        libname : str, optional
            name of library, must be one of ('oseries', 'stresses', 'models'), by
            default "oseries"

        Returns
        -------
        list
            list of items within extent
        """
        xmin, xmax, ymin, ymax = extent
        names = self.conn._parse_names(names, libname)
        if libname == "oseries":
            df = self.oseries.loc[names]
        elif libname == "stresses":
            df = self.stresses.loc[names]
        elif libname == "models":
            onames = np.unique(
                [
                    self.get_models(modelname, return_dict=True)["oseries"]["name"]
                    for modelname in names
                ]
            )
            df = self.oseries.loc[onames]
        else:
            raise ValueError(
                "libname must be one of ['oseries', 'stresses', 'models']"
                f", got '{libname}'"
            )
        mask = (
            (df["x"] <= xmax)
            & (df["x"] >= xmin)
            & (df["y"] >= ymin)
            & (df["y"] <= ymax)
        )
        return df.loc[mask].index.tolist()



# ruff: noqa: D100
from importlib import import_module, metadata
from platform import python_version

import pastas as ps
from packaging.version import parse as parse_version

PASTAS_VERSION = parse_version(ps.__version__)
PASTAS_LEQ_022 = PASTAS_VERSION <= parse_version("0.22.0")
PASTAS_GEQ_150 = PASTAS_VERSION >= parse_version("1.5.0")

__version__ = "1.6.0"


def show_versions(optional=False) -> None:
    """Print the version of dependencies.

    Parameters
    ----------
    optional : bool, optional
        Print the version of optional dependencies, by default False
    """
    msg = (
        f"Pastastore version : {__version__}\n\n"
        f"Python version     : {python_version()}\n"
        f"Pandas version     : {metadata.version('pandas')}\n"
        f"Matplotlib version : {metadata.version('matplotlib')}\n"
        f"Pastas version     : {metadata.version('pastas')}\n"
        f"PyYAML version     : {metadata.version('pyyaml')}\n"
    )
    if optional:
        msg += "\nArcticDB version   : "
        try:
            import_module("arcticdb")
            msg += f"{metadata.version('arctidb')}"
        except ImportError:
            msg += "Not Installed"

    print(msg)



# ruff: noqa: B019
"""Base classes for PastaStore Connectors."""

import functools
import json
import warnings

# import weakref
from abc import ABC, abstractmethod
from collections.abc import Iterable
from itertools import chain
from typing import Dict, List, Optional, Tuple, Union

import pandas as pd
import pastas as ps
from numpy import isin
from packaging.version import parse as parse_version
from pastas.io.pas import PastasEncoder
from tqdm.auto import tqdm

from pastastore.util import ItemInLibraryException, _custom_warning, validate_names
from pastastore.version import PASTAS_GEQ_150, PASTAS_LEQ_022

FrameorSeriesUnion = Union[pd.DataFrame, pd.Series]
warnings.showwarning = _custom_warning


class BaseConnector(ABC):
    """Base Connector class.

    Class holds base logic for dealing with time series and Pastas Models. Create your
    own Connector to a data source by writing a a class that inherits from this
    BaseConnector. Your class has to override each abstractmethod and abstractproperty.
    """

    _default_library_names = [
        "oseries",
        "stresses",
        "models",
        "oseries_models",
    ]

    # whether to check model time series contents against stored copies
    CHECK_MODEL_SERIES_VALUES = True

    # whether to validate time series according to pastas rules
    # True for pastas>=0.23.0 and False for pastas<=0.22.0
    USE_PASTAS_VALIDATE_SERIES = False if PASTAS_LEQ_022 else True

    def __repr__(self):
        """Representation string of the object."""
        return (
            f"<{type(self).__name__}> '{self.name}': "
            f"{self.n_oseries} oseries, "
            f"{self.n_stresses} stresses, "
            f"{self.n_models} models"
        )

    @abstractmethod
    def _get_library(self, libname: str):
        """Get library handle.

        Must be overriden by subclass.

        Parameters
        ----------
        libname : str
            name of the library

        Returns
        -------
        lib : Any
            handle to the library
        """

    @abstractmethod
    def _add_item(
        self,
        libname: str,
        item: Union[FrameorSeriesUnion, Dict],
        name: str,
        metadata: Optional[Dict] = None,
        overwrite: bool = False,
    ) -> None:
        """Add item for both time series and pastas.Models (internal method).

        Must be overriden by subclass.

        Parameters
        ----------
        libname : str
            name of library to add item to
        item : FrameorSeriesUnion or dict
            item to add
        name : str
            name of the item
        metadata : dict, optional
            dictionary containing metadata, by default None
        """

    @abstractmethod
    def _get_item(self, libname: str, name: str) -> Union[FrameorSeriesUnion, Dict]:
        """Get item (series or pastas.Models) (internal method).

        Must be overriden by subclass.

        Parameters
        ----------
        libname : str
            name of library
        name : str
            name of item

        Returns
        -------
        item : FrameorSeriesUnion or dict
            item (time series or pastas.Model)
        """

    @abstractmethod
    def _del_item(self, libname: str, name: str) -> None:
        """Delete items (series or models) (internal method).

        Must be overriden by subclass.

        Parameters
        ----------
        libname : str
            name of library to delete item from
        name : str
            name of item to delete
        """

    @abstractmethod
    def _get_metadata(self, libname: str, name: str) -> Dict:
        """Get metadata (internal method).

        Must be overriden by subclass.

        Parameters
        ----------
        libname : str
            name of the library
        name : str
            name of the item

        Returns
        -------
        metadata : dict
            dictionary containing metadata
        """

    @property
    @abstractmethod
    def oseries_names(self):
        """List of oseries names.

        Property must be overriden by subclass.
        """

    @property
    @abstractmethod
    def stresses_names(self):
        """List of stresses names.

        Property must be overriden by subclass.
        """

    @property
    @abstractmethod
    def model_names(self):
        """List of model names.

        Property must be overriden by subclass.
        """

    def set_check_model_series_values(self, b: bool):
        """Turn CHECK_MODEL_SERIES_VALUES option on (True) or off (False).

        The default option is on (it is highly recommended to keep it that
        way). When turned on, the model time series
        (ml.oseries._series_original, and stressmodel.stress._series_original)
        values are checked against the stored copies in the database. If these
        do not match, an error is raised, and the model is not added to the
        database. This guarantees the stored model will be identical after
        loading from the database. This check is somewhat computationally
        expensive, which is why it can be turned on or off.

        Parameters
        ----------
        b : bool
            boolean indicating whether option should be turned on (True) or
            off (False). Option is on by default.
        """
        self.CHECK_MODEL_SERIES_VALUES = b
        print(f"Model time series checking set to: {b}.")

    def set_use_pastas_validate_series(self, b: bool):
        """Turn USE_PASTAS_VALIDATE_SERIES option on (True) or off (False).

        This will use pastas.validate_oseries() or pastas.validate_stresses()
        to test the time series. If they do not meet the criteria, an error is
        raised. Turning this option off will allow the user to store any time
        series but this will mean that time series models cannot be made from
        stored time series directly and will have to be modified before
        building the models. This in turn will mean that storing the models
        will not work as the stored time series copy is checked against the
        time series in the model to check if they are equal.

        Note: this option requires pastas>=0.23.0, otherwise it is turned off.

        Parameters
        ----------
        b : bool
            boolean indicating whether option should be turned on (True) or
            off (False). Option is on by default.
        """
        self.USE_PASTAS_VALIDATE_SERIES = b
        print(f"Model time series checking set to: {b}.")

    def _pastas_validate(self, validate):
        """Whether to validate time series.

        Parameters
        ----------
        validate : bool, NoneType
            value of validate keyword argument

        Returns
        -------
        b : bool
            return global or local setting (True or False)
        """
        if validate is None:
            return self.USE_PASTAS_VALIDATE_SERIES
        else:
            return validate

    def _add_series(
        self,
        libname: str,
        series: FrameorSeriesUnion,
        name: str,
        metadata: Optional[dict] = None,
        validate: Optional[bool] = None,
        overwrite: bool = False,
    ) -> None:
        """Add series to database (internal method).

        Parameters
        ----------
        libname : str
            name of the library to add the series to
        series : pandas.Series or pandas.DataFrame
            data to add
        name : str
            name of the time series
        metadata : dict, optional
            dictionary containing metadata, by default None
        validate: bool, optional
            use pastas to validate series, default is None, which will use the
            USE_PASTAS_VALIDATE_SERIES value (default is True).
        overwrite : bool, optional
            overwrite existing dataset with the same name,
            by default False

        Raises
        ------
        ItemInLibraryException
            if overwrite is False and name is already in the database
        """
        if not isinstance(name, str):
            name = str(name)
        self._validate_input_series(series)
        series = self._set_series_name(series, name)
        if self._pastas_validate(validate):
            if libname == "oseries":
                if PASTAS_GEQ_150 and not ps.validate_oseries(series):
                    raise ValueError(
                        "oseries does not meet pastas criteria,"
                        " see `ps.validate_oseries()`!"
                    )
                else:
                    ps.validate_oseries(series)
            else:
                if PASTAS_GEQ_150 and not ps.validate_stress(series):
                    raise ValueError(
                        "stress does not meet pastas criteria,"
                        " see `ps.validate_stress()`!"
                    )
                else:
                    ps.validate_stress(series)
        in_store = getattr(self, f"{libname}_names")
        if name not in in_store or overwrite:
            self._add_item(
                libname, series, name, metadata=metadata, overwrite=overwrite
            )
            self._clear_cache(libname)
        else:
            raise ItemInLibraryException(
                f"Item with name '{name}' already" f" in '{libname}' library!"
            )

    def _update_series(
        self,
        libname: str,
        series: FrameorSeriesUnion,
        name: str,
        metadata: Optional[dict] = None,
        validate: Optional[bool] = None,
    ) -> None:
        """Update time series (internal method).

        Parameters
        ----------
        libname : str
            name of library
        series : FrameorSeriesUnion
            time series containing update values
        name : str
            name of the time series to update
        metadata : Optional[dict], optional
            optionally provide metadata dictionary which will also update
            the current stored metadata dictionary, by default None
        validate: bool, optional
            use pastas to validate series, default is None, which will use the
            USE_PASTAS_VALIDATE_SERIES value (default is True).
        """
        if libname not in ["oseries", "stresses"]:
            raise ValueError("Library must be 'oseries' or 'stresses'!")
        self._validate_input_series(series)
        series = self._set_series_name(series, name)
        stored = self._get_series(libname, name, progressbar=False)
        if self.conn_type == "pas" and not isinstance(series, type(stored)):
            if isinstance(series, pd.DataFrame):
                stored = stored.to_frame()
        # get union of index
        idx_union = stored.index.union(series.index)
        # update series with new values
        update = stored.reindex(idx_union)
        update.update(series)
        # metadata
        update_meta = self._get_metadata(libname, name)
        if metadata is not None:
            update_meta.update(metadata)
        self._add_series(
            libname,
            update,
            name,
            metadata=update_meta,
            validate=validate,
            overwrite=True,
        )

    def _upsert_series(
        self,
        libname: str,
        series: FrameorSeriesUnion,
        name: str,
        metadata: Optional[dict] = None,
        validate: Optional[bool] = None,
    ) -> None:
        """Update or insert series depending on whether it exists in store.

        Parameters
        ----------
        libname : str
            name of library
        series : FrameorSeriesUnion
            time series to update/insert
        name : str
            name of the time series
        metadata : Optional[dict], optional
            metadata dictionary, by default None
        validate : bool, optional
            use pastas to validate series, default is None, which will use the
            USE_PASTAS_VALIDATE_SERIES value (default is True).
        """
        if libname not in ["oseries", "stresses"]:
            raise ValueError("Library must be 'oseries' or 'stresses'!")
        if name in getattr(self, f"{libname}_names"):
            self._update_series(
                libname, series, name, metadata=metadata, validate=validate
            )
        else:
            self._add_series(
                libname, series, name, metadata=metadata, validate=validate
            )

    def update_metadata(self, libname: str, name: str, metadata: dict) -> None:
        """Update metadata.

        Note: also retrieves and stores time series as updating only metadata
        is not supported for some Connectors.

        Parameters
        ----------
        libname : str
            name of library
        name : str
            name of the item for which to update metadata
        metadata : dict
            metadata dictionary that will be used to update the stored
            metadata
        """
        if libname not in ["oseries", "stresses"]:
            raise ValueError("Library must be 'oseries' or 'stresses'!")
        update_meta = self._get_metadata(libname, name)
        update_meta.update(metadata)
        # get series, since just updating metadata is not really defined
        # in all cases
        s = self._get_series(libname, name, progressbar=False)
        self._add_series(libname, s, name, metadata=update_meta, overwrite=True)

    def add_oseries(
        self,
        series: FrameorSeriesUnion,
        name: str,
        metadata: Optional[dict] = None,
        validate: Optional[bool] = None,
        overwrite: bool = False,
    ) -> None:
        """Add oseries to the database.

        Parameters
        ----------
        series : pandas.Series or pandas.DataFrame
            data to add
        name : str
            name of the time series
        metadata : dict, optional
            dictionary containing metadata, by default None.
        validate : bool, optional
            use pastas to validate series, default is None, which will use the
            USE_PASTAS_VALIDATE_SERIES value (default is True).
        overwrite : bool, optional
            overwrite existing dataset with the same name,
            by default False
        """
        series, metadata = self._parse_series_input(series, metadata)
        self._add_series(
            "oseries",
            series,
            name=name,
            metadata=metadata,
            validate=validate,
            overwrite=overwrite,
        )

    def add_stress(
        self,
        series: FrameorSeriesUnion,
        name: str,
        kind: str,
        metadata: Optional[dict] = None,
        validate: Optional[bool] = None,
        overwrite: bool = False,
    ) -> None:
        """Add stress to the database.

        Parameters
        ----------
        series : pandas.Series or pandas.DataFrame
            data to add, if pastas.Timeseries is passed, series_orignal
            and metadata is stored in database
        name : str
            name of the time series
        kind : str
            category to identify type of stress, this label is added to the
            metadata dictionary.
        metadata : dict, optional
            dictionary containing metadata, by default None.
        validate : bool, optional
            use pastas to validate series, default is True
        overwrite : bool, optional
            overwrite existing dataset with the same name,
            by default False
        """
        series, metadata = self._parse_series_input(series, metadata)
        if metadata is None:
            metadata = {}
        metadata["kind"] = kind
        self._add_series(
            "stresses",
            series,
            name=name,
            metadata=metadata,
            validate=validate,
            overwrite=overwrite,
        )

    def add_model(
        self,
        ml: Union[ps.Model, dict],
        overwrite: bool = False,
        validate_metadata: bool = False,
    ) -> None:
        """Add model to the database.

        Parameters
        ----------
        ml : pastas.Model or dict
            pastas Model or dictionary to add to the database
        overwrite : bool, optional
            if True, overwrite existing model, by default False
        validate_metadata, bool optional
            remove unsupported characters from metadata dictionary keys

        Raises
        ------
        TypeError
            if model is not pastas.Model or dict
        ItemInLibraryException
            if overwrite is False and model is already in the database
        """
        if isinstance(ml, ps.Model):
            mldict = ml.to_dict(series=False)
            name = ml.name
            if validate_metadata:
                metadata = validate_names(d=ml.oseries.metadata)
            else:
                metadata = ml.oseries.metadata
        elif isinstance(ml, dict):
            mldict = ml
            name = ml["name"]
            metadata = None
        else:
            raise TypeError("Expected pastas.Model or dict!")
        if not isinstance(name, str):
            name = str(name)
        if name not in self.model_names or overwrite:
            # check if stressmodels supported
            self._check_stressmodels_supported(ml)
            # check if oseries and stresses exist in store
            self._check_model_series_names_for_store(ml)
            self._check_oseries_in_store(ml)
            self._check_stresses_in_store(ml)
            # write model to store
            self._add_item(
                "models", mldict, name, metadata=metadata, overwrite=overwrite
            )
        else:
            raise ItemInLibraryException(
                f"Model with name '{name}' " "already in 'models' library!"
            )
        self._clear_cache("_modelnames_cache")
        self._add_oseries_model_links(str(mldict["oseries"]["name"]), name)

    @staticmethod
    def _parse_series_input(
        series: FrameorSeriesUnion,
        metadata: Optional[Dict] = None,
    ) -> Tuple[FrameorSeriesUnion, Optional[Dict]]:
        """Parse series input (internal method).

        Parameters
        ----------
        series : FrameorSeriesUnion,
            series object to parse
        metadata : dict, optional
            metadata dictionary or None, by default None

        Returns
        -------
        series, metadata : FrameorSeriesUnion, Optional[Dict]
            time series as pandas.Series or DataFrame and optionally
            metadata dictionary
        """
        if isinstance(series, ps.timeseries.TimeSeries):
            raise DeprecationWarning(
                "Pastas TimeSeries objects are no longer supported!"
            )
        s = series
        m = metadata
        return s, m

    def update_oseries(
        self,
        series: FrameorSeriesUnion,
        name: str,
        metadata: Optional[dict] = None,
    ) -> None:
        """Update oseries values.

        Parameters
        ----------
        series : FrameorSeriesUnion
            time series to update stored oseries with
        name : str
            name of the oseries to update
        metadata : Optional[dict], optional
            optionally provide metadata, which will update
            the stored metadata dictionary, by default None
        """
        series, metadata = self._parse_series_input(series, metadata)
        self._update_series("oseries", series, name, metadata=metadata)

    def upsert_oseries(
        self,
        series: FrameorSeriesUnion,
        name: str,
        metadata: Optional[dict] = None,
    ) -> None:
        """Update or insert oseries values depending on whether it exists.

        Parameters
        ----------
        series : FrameorSeriesUnion
            time series to update/insert
        name : str
            name of the oseries
        metadata : Optional[dict], optional
            optionally provide metadata, which will update
            the stored metadata dictionary if it exists, by default None
        """
        series, metadata = self._parse_series_input(series, metadata)
        self._upsert_series("oseries", series, name, metadata=metadata)

    def update_stress(
        self,
        series: FrameorSeriesUnion,
        name: str,
        metadata: Optional[dict] = None,
    ) -> None:
        """Update stresses values.

        Note: the 'kind' attribute of a stress cannot be updated! To update
        the 'kind' delete and add the stress again.

        Parameters
        ----------
        series : FrameorSeriesUnion
            time series to update stored stress with
        name : str
            name of the stress to update
        metadata : Optional[dict], optional
            optionally provide metadata, which will update
            the stored metadata dictionary, by default None
        """
        series, metadata = self._parse_series_input(series, metadata)
        self._update_series("stresses", series, name, metadata=metadata)

    def upsert_stress(
        self,
        series: FrameorSeriesUnion,
        name: str,
        kind: str,
        metadata: Optional[dict] = None,
    ) -> None:
        """Update or insert stress values depending on whether it exists.

        Parameters
        ----------
        series : FrameorSeriesUnion
            time series to update/insert
        name : str
            name of the stress
        metadata : Optional[dict], optional
            optionally provide metadata, which will update
            the stored metadata dictionary if it exists, by default None
        """
        series, metadata = self._parse_series_input(series, metadata)
        if metadata is None:
            metadata = {}
        metadata["kind"] = kind
        self._upsert_series("stresses", series, name, metadata=metadata)

    def del_models(self, names: Union[list, str]) -> None:
        """Delete model(s) from the database.

        Parameters
        ----------
        names : str or list of str
            name(s) of the model to delete
        """
        for n in self._parse_names(names, libname="models"):
            mldict = self.get_models(n, return_dict=True)
            oname = mldict["oseries"]["name"]
            self._del_item("models", n)
            self._del_oseries_model_link(oname, n)
        self._clear_cache("_modelnames_cache")

    def del_model(self, names: Union[list, str]) -> None:
        """Delete model(s) from the database.

        Alias for del_models().

        Parameters
        ----------
        names : str or list of str
            name(s) of the model to delete
        """
        self.del_models(names=names)

    def del_oseries(self, names: Union[list, str], remove_models: bool = False):
        """Delete oseries from the database.

        Parameters
        ----------
        names : str or list of str
            name(s) of the oseries to delete
        remove_models : bool, optional
            also delete models for deleted oseries, default is False
        """
        names = self._parse_names(names, libname="oseries")
        for n in names:
            self._del_item("oseries", n)
        self._clear_cache("oseries")
        # remove associated models from database
        if remove_models:
            modelnames = list(
                chain.from_iterable([self.oseries_models.get(n, []) for n in names])
            )
            self.del_models(modelnames)

    def del_stress(self, names: Union[list, str]):
        """Delete stress from the database.

        Parameters
        ----------
        names : str or list of str
            name(s) of the stress to delete
        """
        for n in self._parse_names(names, libname="stresses"):
            self._del_item("stresses", n)
        self._clear_cache("stresses")

    def _get_series(
        self,
        libname: str,
        names: Union[list, str],
        progressbar: bool = True,
        squeeze: bool = True,
    ) -> FrameorSeriesUnion:
        """Get time series (internal method).

        Parameters
        ----------
        libname : str
            name of the library
        names : str or list of str
            names of the time series to load
        progressbar : bool, optional
            show progressbar, by default True
        squeeze : bool, optional
            if True return DataFrame or Series instead of dictionary
            for single entry

        Returns
        -------
        pandas.DataFrame or dict of pandas.DataFrames
            either returns time series as pandas.DataFrame or
            dictionary containing the time series.
        """
        ts = {}
        names = self._parse_names(names, libname=libname)
        desc = f"Get {libname}"
        for n in tqdm(names, desc=desc) if progressbar else names:
            ts[n] = self._get_item(libname, n)
        # return frame if len == 1
        if len(ts) == 1 and squeeze:
            return ts[n]
        else:
            return ts

    def get_metadata(
        self,
        libname: str,
        names: Union[list, str],
        progressbar: bool = False,
        as_frame: bool = True,
        squeeze: bool = True,
    ) -> Union[dict, pd.DataFrame]:
        """Read metadata from database.

        Parameters
        ----------
        libname : str
            name of the library containing the dataset
        names : str or list of str
            names of the datasets for which to read the metadata
        squeeze : bool, optional
            if True return dict instead of list of dict
            for single entry

        Returns
        -------
        dict or pandas.DataFrame
            returns metadata dictionary or DataFrame of metadata
        """
        metalist = []
        names = self._parse_names(names, libname=libname)
        desc = f"Get metadata {libname}"
        for n in tqdm(names, desc=desc) if progressbar else names:
            imeta = self._get_metadata(libname, n)
            if imeta is None:
                imeta = {}
            metalist.append(imeta)
        if as_frame:
            meta = self._meta_list_to_frame(metalist, names=names)
            return meta
        else:
            if len(metalist) == 1 and squeeze:
                return metalist[0]
            else:
                return metalist

    def get_oseries(
        self,
        names: Union[list, str],
        return_metadata: bool = False,
        progressbar: bool = False,
        squeeze: bool = True,
    ) -> Union[Union[FrameorSeriesUnion, Dict], Optional[Union[Dict, List]]]:
        """Get oseries from database.

        Parameters
        ----------
        names : str or list of str
            names of the oseries to load
        return_metadata : bool, optional
            return metadata as dictionary or list of dictionaries,
            default is False
        progressbar : bool, optional
            show progressbar, by default False
        squeeze : bool, optional
            if True return DataFrame or Series instead of dictionary
            for single entry

        Returns
        -------
        oseries : pandas.DataFrame or dict of DataFrames
            returns time series as DataFrame or dictionary of DataFrames if
            multiple names were passed
        metadata : dict or list of dict
            metadata for each oseries, only returned if return_metadata=True
        """
        oseries = self._get_series(
            "oseries", names, progressbar=progressbar, squeeze=squeeze
        )
        if return_metadata:
            metadata = self.get_metadata(
                "oseries",
                names,
                progressbar=progressbar,
                as_frame=False,
                squeeze=squeeze,
            )
            return oseries, metadata
        else:
            return oseries

    def get_stresses(
        self,
        names: Union[list, str],
        return_metadata: bool = False,
        progressbar: bool = False,
        squeeze: bool = True,
    ) -> Union[Union[FrameorSeriesUnion, Dict], Optional[Union[Dict, List]]]:
        """Get stresses from database.

        Parameters
        ----------
        names : str or list of str
            names of the stresses to load
        return_metadata : bool, optional
            return metadata as dictionary or list of dictionaries,
            default is False
        progressbar : bool, optional
            show progressbar, by default False
        squeeze : bool, optional
            if True return DataFrame or Series instead of dictionary
            for single entry

        Returns
        -------
        stresses : pandas.DataFrame or dict of DataFrames
            returns time series as DataFrame or dictionary of DataFrames if
            multiple names were passed
        metadata : dict or list of dict
            metadata for each stress, only returned if return_metadata=True
        """
        stresses = self._get_series(
            "stresses", names, progressbar=progressbar, squeeze=squeeze
        )
        if return_metadata:
            metadata = self.get_metadata(
                "stresses",
                names,
                progressbar=progressbar,
                as_frame=False,
                squeeze=squeeze,
            )
            return stresses, metadata
        else:
            return stresses

    def get_stress(
        self,
        names: Union[list, str],
        return_metadata: bool = False,
        progressbar: bool = False,
        squeeze: bool = True,
    ) -> Union[Union[FrameorSeriesUnion, Dict], Optional[Union[Dict, List]]]:
        """Get stresses from database.

        Alias for `get_stresses()`

        Parameters
        ----------
        names : str or list of str
            names of the stresses to load
        return_metadata : bool, optional
            return metadata as dictionary or list of dictionaries,
            default is False
        progressbar : bool, optional
            show progressbar, by default False
        squeeze : bool, optional
            if True return DataFrame or Series instead of dictionary
            for single entry

        Returns
        -------
        stresses : pandas.DataFrame or dict of DataFrames
            returns time series as DataFrame or dictionary of DataFrames if
            multiple names were passed
        metadata : dict or list of dict
            metadata for each stress, only returned if return_metadata=True
        """
        return self.get_stresses(
            names,
            return_metadata=return_metadata,
            progressbar=progressbar,
            squeeze=squeeze,
        )

    def get_models(
        self,
        names: Union[list, str],
        return_dict: bool = False,
        progressbar: bool = False,
        squeeze: bool = True,
        update_ts_settings: bool = False,
    ) -> Union[ps.Model, list]:
        """Load models from database.

        Parameters
        ----------
        names : str or list of str
            names of the models to load
        return_dict : bool, optional
            return model dictionary instead of pastas.Model (much
            faster for obtaining parameters, for example)
        progressbar : bool, optional
            show progressbar, by default False
        squeeze : bool, optional
            if True return Model instead of list of Models
            for single entry
        update_ts_settings : bool, optional
            update time series settings based on time series in store.
            overwrites stored tmin/tmax in model.

        Returns
        -------
        pastas.Model or list of pastas.Model
            return pastas model, or list of models if multiple names were
            passed
        """
        models = []
        names = self._parse_names(names, libname="models")
        desc = "Get models"
        for n in tqdm(names, desc=desc) if progressbar else names:
            data = self._get_item("models", n)
            if return_dict:
                ml = data
            else:
                ml = self._parse_model_dict(data, update_ts_settings=update_ts_settings)
            models.append(ml)
        if len(models) == 1 and squeeze:
            return models[0]
        else:
            return models

    def get_model(
        self,
        names: Union[list, str],
        return_dict: bool = False,
        progressbar: bool = False,
        squeeze: bool = True,
        update_ts_settings: bool = False,
    ) -> Union[ps.Model, list]:
        """Load models from database.

        Alias for get_models().

        Parameters
        ----------
        names : str or list of str
            names of the models to load
        return_dict : bool, optional
            return model dictionary instead of pastas.Model (much
            faster for obtaining parameters, for example)
        progressbar : bool, optional
            show progressbar, by default False
        squeeze : bool, optional
            if True return Model instead of list of Models
            for single entry
        update_ts_settings : bool, optional
            update time series settings based on time series in store.
            overwrites stored tmin/tmax in model.

        Returns
        -------
        pastas.Model or list of pastas.Model
            return pastas model, or list of models if multiple names were
            passed
        """
        return self.get_models(
            names,
            return_dict=return_dict,
            progressbar=progressbar,
            squeeze=squeeze,
            update_ts_settings=update_ts_settings,
        )

    def empty_library(
        self, libname: str, prompt: bool = True, progressbar: bool = True
    ):
        """Empty library of all its contents.

        Parameters
        ----------
        libname : str
            name of the library
        prompt : bool, optional
            prompt user for input before deleting
            contents, by default True. Default answer is
            "n", user must enter 'y' to delete contents
        progressbar : bool, optional
            show progressbar, by default True
        """
        if prompt:
            ui = input(
                f"Do you want to empty '{libname}'"
                " library of all its contents? [y/N] "
            )
            if ui.lower() != "y":
                return

        if libname == "models":
            # also delete linked modelnames linked to oseries
            libs = ["models", "oseries_models"]
        else:
            libs = [libname]

        # delete items and clear caches
        for libname in libs:
            names = self._parse_names(None, libname)
            for name in (
                tqdm(names, desc=f"Deleting items from {libname}")
                if progressbar
                else names
            ):
                self._del_item(libname, name)
            self._clear_cache(libname)
            print(f"Emptied library {libname} in {self.name}: " f"{self.__class__}")

    def _iter_series(self, libname: str, names: Optional[List[str]] = None):
        """Iterate over time series in library (internal method).

        Parameters
        ----------
        libname : str
            name of library (e.g. 'oseries' or 'stresses')
        names : Optional[List[str]], optional
            list of names, by default None, which defaults to
            all stored series


        Yields
        ------
        pandas.Series or pandas.DataFrame
            time series contained in library
        """
        names = self._parse_names(names, libname)
        for nam in names:
            yield self._get_series(libname, nam, progressbar=False)

    def iter_oseries(self, names: Optional[List[str]] = None):
        """Iterate over oseries in library.

        Parameters
        ----------
        names : Optional[List[str]], optional
            list of oseries names, by default None, which defaults to
            all stored series


        Yields
        ------
        pandas.Series or pandas.DataFrame
            oseries contained in library
        """
        yield from self._iter_series("oseries", names=names)

    def iter_stresses(self, names: Optional[List[str]] = None):
        """Iterate over stresses in library.

        Parameters
        ----------
        names : Optional[List[str]], optional
            list of stresses names, by default None, which defaults to
            all stored series


        Yields
        ------
        pandas.Series or pandas.DataFrame
            stresses contained in library
        """
        yield from self._iter_series("stresses", names=names)

    def iter_models(
        self, modelnames: Optional[List[str]] = None, return_dict: bool = False
    ):
        """Iterate over models in library.

        Parameters
        ----------
        modelnames : Optional[List[str]], optional
            list of models to iterate over, by default None which uses
            all models
        return_dict : bool, optional
            if True, return model as dictionary, by default False,
            which returns a pastas.Model.

        Yields
        ------
        pastas.Model or dict
            time series model
        """
        modelnames = self._parse_names(modelnames, "models")
        for mlnam in modelnames:
            yield self.get_models(mlnam, return_dict=return_dict, progressbar=False)

    def _add_oseries_model_links(self, onam: str, mlnames: Union[str, List[str]]):
        """Add model name to stored list of models per oseries.

        Parameters
        ----------
        onam : str
            name of oseries
        mlnames : Union[str, List[str]]
            model name or list of model names for an oseries with name
            onam.
        """
        # get stored list of model names
        if str(onam) in self.oseries_with_models:
            modellist = self._get_item("oseries_models", onam)
        else:
            # else empty list
            modellist = []
        # if one model name, make list for loop
        if isinstance(mlnames, str):
            mlnames = [mlnames]
        # loop over model names
        for iml in mlnames:
            # if not present, add to list
            if iml not in modellist:
                modellist.append(iml)
        self._add_item("oseries_models", modellist, onam, overwrite=True)
        self._clear_cache("oseries_models")

    def _del_oseries_model_link(self, onam, mlnam):
        """Delete model name from stored list of models per oseries.

        Parameters
        ----------
        onam : str
            name of oseries
        mlnam : str
            name of model
        """
        modellist = self._get_item("oseries_models", onam)
        modellist.remove(mlnam)
        if len(modellist) == 0:
            self._del_item("oseries_models", onam)
        else:
            self._add_item("oseries_models", modellist, onam, overwrite=True)
        self._clear_cache("oseries_models")

    def _update_all_oseries_model_links(self):
        """Add all model names to oseries metadata dictionaries.

        Used for old PastaStore versions, where relationship between oseries and models
        was not stored. If there are any models in the database and if the
        oseries_models library is empty, loops through all models to determine which
        oseries each model belongs to.
        """
        # get oseries_models library if there are any contents, if empty
        # add all model links.
        if self.n_models > 0:
            if len(self.oseries_models) == 0:
                links = self._get_all_oseries_model_links()
                for onam, mllinks in tqdm(
                    links.items(),
                    desc="Store models per oseries",
                    total=len(links),
                ):
                    self._add_oseries_model_links(onam, mllinks)

    def _get_all_oseries_model_links(self):
        """Get all model names per oseries in dictionary.

        Returns
        -------
        links : dict
            dictionary with oseries names as keys and lists of model names as
            values
        """
        links = {}
        for mldict in tqdm(
            self.iter_models(return_dict=True),
            total=self.n_models,
            desc="Get models per oseries",
        ):
            onam = mldict["oseries"]["name"]
            mlnam = mldict["name"]
            if onam in links:
                links[onam].append(mlnam)
            else:
                links[onam] = [mlnam]
        return links

    @staticmethod
    def _clear_cache(libname: str) -> None:
        """Clear cached property."""
        if libname == "models":
            libname = "_modelnames_cache"
        getattr(BaseConnector, libname).fget.cache_clear()

    @property  # type: ignore
    @functools.lru_cache()
    def oseries(self):
        """Dataframe with overview of oseries."""
        return self.get_metadata("oseries", self.oseries_names)

    @property  # type: ignore
    @functools.lru_cache()
    def stresses(self):
        """Dataframe with overview of stresses."""
        return self.get_metadata("stresses", self.stresses_names)

    @property  # type: ignore
    @functools.lru_cache()
    def _modelnames_cache(self):
        """List of model names."""
        return self.model_names

    @property
    def n_oseries(self):
        """
        Returns the number of oseries.

        Returns
        -------
        int
            The number of oseries names.
        """
        return len(self.oseries_names)

    @property
    def n_stresses(self):
        """
        Returns the number of stresses.

        Returns
        -------
        int
            The number of stresses.
        """
        return len(self.stresses_names)

    @property
    def n_models(self):
        """
        Returns the number of models in the store.

        Returns
        -------
            int
                The number of models in the store.
        """
        return len(self.model_names)

    @property  # type: ignore
    @functools.lru_cache()
    def oseries_models(self):
        """List of model names per oseries.

        Returns
        -------
        d : dict
            dictionary with oseries names as keys and list of model names as
            values
        """
        d = {}
        for onam in self.oseries_with_models:
            d[onam] = self._get_item("oseries_models", onam)
        return d


class ConnectorUtil:
    """Mix-in class for general Connector helper functions.

    Only for internal methods, and not methods that are related to CRUD operations on
    database.
    """

    def _parse_names(
        self,
        names: Optional[Union[list, str]] = None,
        libname: Optional[str] = "oseries",
    ) -> list:
        """Parse names kwarg, returns iterable with name(s) (internal method).

        Parameters
        ----------
        names : Union[list, str], optional
            str or list of str or None or 'all' (last two options
            retrieves all names)
        libname : str, optional
            name of library, default is 'oseries'

        Returns
        -------
        list
            list of names
        """
        if not isinstance(names, str) and isinstance(names, Iterable):
            return names
        elif isinstance(names, str) and names != "all":
            return [names]
        elif names is None or names == "all":
            if libname == "oseries":
                return self.oseries_names
            elif libname == "stresses":
                return self.stresses_names
            elif libname == "models":
                return self.model_names
            elif libname == "oseries_models":
                return self.oseries_with_models
            else:
                raise ValueError(f"No library '{libname}'!")
        else:
            raise NotImplementedError(f"Cannot parse 'names': {names}")

    @staticmethod
    def _meta_list_to_frame(metalist: list, names: list):
        """Convert list of metadata dictionaries to DataFrame.

        Parameters
        ----------
        metalist : list
            list of metadata dictionaries
        names : list
            list of names corresponding to data in metalist

        Returns
        -------
        pandas.DataFrame
            DataFrame containing overview of metadata
        """
        # convert to dataframe
        if len(metalist) > 1:
            meta = pd.DataFrame(metalist)
            if len({"x", "y"}.difference(meta.columns)) == 0:
                meta["x"] = meta["x"].astype(float)
                meta["y"] = meta["y"].astype(float)
        elif len(metalist) == 1:
            meta = pd.DataFrame(metalist)
        elif len(metalist) == 0:
            meta = pd.DataFrame()

        meta.index = names
        meta.index.name = "name"
        return meta

    def _parse_model_dict(self, mdict: dict, update_ts_settings: bool = False):
        """Parse dictionary describing pastas models (internal method).

        Parameters
        ----------
        mdict : dict
            dictionary describing pastas.Model
        update_ts_settings : bool, optional
            update stored tmin and tmax in time series settings
            based on time series loaded from store.

        Returns
        -------
        ml : pastas.Model
            time series analysis model
        """
        PASFILE_LEQ_022 = parse_version(
            mdict["file_info"]["pastas_version"]
        ) <= parse_version("0.22.0")

        # oseries
        if "series" not in mdict["oseries"]:
            name = str(mdict["oseries"]["name"])
            if name not in self.oseries.index:
                msg = "oseries '{}' not present in library".format(name)
                raise LookupError(msg)
            mdict["oseries"]["series"] = self.get_oseries(name).squeeze()
            # update tmin/tmax from time series
            if update_ts_settings:
                mdict["oseries"]["settings"]["tmin"] = mdict["oseries"]["series"].index[
                    0
                ]
                mdict["oseries"]["settings"]["tmax"] = mdict["oseries"]["series"].index[
                    -1
                ]

        # StressModel, WellModel
        for ts in mdict["stressmodels"].values():
            if "stress" in ts.keys():
                # WellModel
                classkey = "stressmodel" if PASFILE_LEQ_022 else "class"
                if ts[classkey] == "WellModel":
                    for stress in ts["stress"]:
                        if "series" not in stress:
                            name = str(stress["name"])
                            if name in self.stresses.index:
                                stress["series"] = self.get_stresses(name).squeeze()
                                # update tmin/tmax from time series
                                if update_ts_settings:
                                    stress["settings"]["tmin"] = stress["series"].index[
                                        0
                                    ]
                                    stress["settings"]["tmax"] = stress["series"].index[
                                        -1
                                    ]
                # StressModel
                else:
                    for stress in ts["stress"] if PASFILE_LEQ_022 else [ts["stress"]]:
                        if "series" not in stress:
                            name = str(stress["name"])
                            if name in self.stresses.index:
                                stress["series"] = self.get_stresses(name).squeeze()
                                # update tmin/tmax from time series
                                if update_ts_settings:
                                    stress["settings"]["tmin"] = stress["series"].index[
                                        0
                                    ]
                                    stress["settings"]["tmax"] = stress["series"].index[
                                        -1
                                    ]

            # RechargeModel, TarsoModel
            if ("prec" in ts.keys()) and ("evap" in ts.keys()):
                for stress in [ts["prec"], ts["evap"]]:
                    if "series" not in stress:
                        name = str(stress["name"])
                        if name in self.stresses.index:
                            stress["series"] = self.get_stresses(name).squeeze()
                            # update tmin/tmax from time series
                            if update_ts_settings:
                                stress["settings"]["tmin"] = stress["series"].index[0]
                                stress["settings"]["tmax"] = stress["series"].index[-1]
                        else:
                            msg = "stress '{}' not present in library".format(name)
                            raise KeyError(msg)

        # hack for pcov w dtype object (when filled with NaNs on store?)
        if "fit" in mdict:
            if "pcov" in mdict["fit"]:
                pcov = mdict["fit"]["pcov"]
                if pcov.dtypes.apply(lambda dtyp: isinstance(dtyp, object)).any():
                    mdict["fit"]["pcov"] = pcov.astype(float)

        # check pastas version vs pas-file version
        file_version = mdict["file_info"]["pastas_version"]

        # check file version and pastas version
        # if file<0.23  and pastas>=1.0 --> error
        PASTAS_GT_023 = parse_version(ps.__version__) > parse_version("0.23.1")
        if PASFILE_LEQ_022 and PASTAS_GT_023:
            raise UserWarning(
                f"This file was created with Pastas v{file_version} "
                f"and cannot be loaded with Pastas v{ps.__version__} Please load and "
                "save the file with Pastas 0.23 first to update the file "
                "format."
            )

        try:
            # pastas>=0.15.0
            ml = ps.io.base._load_model(mdict)
        except AttributeError:
            # pastas<0.15.0
            ml = ps.io.base.load_model(mdict)
        return ml

    @staticmethod
    def _validate_input_series(series):
        """Check if series is pandas.DataFrame or pandas.Series.

        Parameters
        ----------
        series : object
            object to validate

        Raises
        ------
        TypeError
            if object is not of type pandas.DataFrame or pandas.Series
        """
        if not (isinstance(series, pd.DataFrame) or isinstance(series, pd.Series)):
            raise TypeError("Please provide pandas.DataFrame or pandas.Series!")
        if isinstance(series, pd.DataFrame):
            if series.columns.size > 1:
                raise ValueError("Only DataFrames with one column are supported!")

    @staticmethod
    def _set_series_name(series, name):
        """Set series name to match user defined name in store.

        Parameters
        ----------
        series : pandas.Series or pandas.DataFrame
            set name for this time series
        name : str
            name of the time series (used in the pastastore)
        """
        if isinstance(series, pd.Series):
            series.name = name
            # empty string on index name causes trouble when reading
            # data from ArcticDB: TODO: check if still an issue?
            if series.index.name == "":
                series.index.name = None

        if isinstance(series, pd.DataFrame):
            series.columns = [name]
            # check for hydropandas objects which are instances of DataFrame but
            # do have a name attribute
            if hasattr(series, "name"):
                series.name = name
        return series

    @staticmethod
    def _check_stressmodels_supported(ml):
        supported_stressmodels = [
            "StressModel",
            "StressModel2",
            "RechargeModel",
            "WellModel",
            "TarsoModel",
            "Constant",
            "LinearTrend",
            "StepModel",
        ]
        if isinstance(ml, ps.Model):
            smtyps = [sm._name for sm in ml.stressmodels.values()]
        elif isinstance(ml, dict):
            classkey = "stressmodel" if PASTAS_LEQ_022 else "class"
            smtyps = [sm[classkey] for sm in ml["stressmodels"].values()]
        check = isin(smtyps, supported_stressmodels)
        if not all(check):
            unsupported = set(smtyps) - set(supported_stressmodels)
            raise NotImplementedError(
                "PastaStore does not support storing models with the "
                f"following stressmodels: {unsupported}"
            )

    @staticmethod
    def _check_model_series_names_for_store(ml):
        prec_evap_model = ["RechargeModel", "TarsoModel"]

        if isinstance(ml, ps.Model):
            series_names = [
                istress.series.name
                for sm in ml.stressmodels.values()
                for istress in sm.stress
            ]

        elif isinstance(ml, dict):
            # non RechargeModel, Tarsomodel, WellModel stressmodels
            classkey = "stressmodel" if PASTAS_LEQ_022 else "class"
            if PASTAS_LEQ_022:
                series_names = [
                    istress["name"]
                    for sm in ml["stressmodels"].values()
                    if sm[classkey] not in (prec_evap_model + ["WellModel"])
                    for istress in sm["stress"]
                ]
            else:
                series_names = [
                    sm["stress"]["name"]
                    for sm in ml["stressmodels"].values()
                    if sm[classkey] not in (prec_evap_model + ["WellModel"])
                ]

            # WellModel
            if isin(
                ["WellModel"],
                [i[classkey] for i in ml["stressmodels"].values()],
            ).any():
                series_names += [
                    istress["name"]
                    for sm in ml["stressmodels"].values()
                    if sm[classkey] in ["WellModel"]
                    for istress in sm["stress"]
                ]

            # RechargeModel, TarsoModel
            if isin(
                prec_evap_model,
                [i[classkey] for i in ml["stressmodels"].values()],
            ).any():
                series_names += [
                    istress["name"]
                    for sm in ml["stressmodels"].values()
                    if sm[classkey] in prec_evap_model
                    for istress in [sm["prec"], sm["evap"]]
                ]

        else:
            raise TypeError("Expected pastas.Model or dict!")
        if len(series_names) - len(set(series_names)) > 0:
            msg = (
                "There are multiple stresses series with the same name! "
                "Each series name must be unique for the PastaStore!"
            )
            raise ValueError(msg)

    def _check_oseries_in_store(self, ml: Union[ps.Model, dict]):
        """Check if Model oseries are contained in PastaStore (internal method).

        Parameters
        ----------
        ml : Union[ps.Model, dict]
            pastas Model
        """
        if isinstance(ml, ps.Model):
            name = ml.oseries.name
        elif isinstance(ml, dict):
            name = str(ml["oseries"]["name"])
        else:
            raise TypeError("Expected pastas.Model or dict!")
        if name not in self.oseries.index:
            msg = (
                f"Cannot add model because oseries '{name}' "
                "is not contained in store."
            )
            raise LookupError(msg)
        # expensive check
        if self.CHECK_MODEL_SERIES_VALUES and isinstance(ml, ps.Model):
            s_org = self.get_oseries(name).squeeze().dropna()
            if PASTAS_LEQ_022:
                so = ml.oseries.series_original
            else:
                so = ml.oseries._series_original
            if not so.dropna().equals(s_org):
                raise ValueError(
                    f"Cannot add model because model oseries '{name}'"
                    " is different from stored oseries!"
                )

    def _check_stresses_in_store(self, ml: Union[ps.Model, dict]):
        """Check if stresses time series are contained in PastaStore (internal method).

        Parameters
        ----------
        ml : Union[ps.Model, dict]
            pastas Model
        """
        prec_evap_model = ["RechargeModel", "TarsoModel"]
        if isinstance(ml, ps.Model):
            for sm in ml.stressmodels.values():
                if sm._name in prec_evap_model:
                    stresses = [sm.prec, sm.evap]
                else:
                    stresses = sm.stress
                for s in stresses:
                    if str(s.name) not in self.stresses.index:
                        msg = (
                            f"Cannot add model because stress '{s.name}' "
                            "is not contained in store."
                        )
                        raise LookupError(msg)
                    if self.CHECK_MODEL_SERIES_VALUES:
                        s_org = self.get_stresses(s.name).squeeze()
                        if PASTAS_LEQ_022:
                            so = s.series_original
                        else:
                            so = s._series_original
                        if not so.equals(s_org):
                            raise ValueError(
                                f"Cannot add model because model stress "
                                f"'{s.name}' is different from stored stress!"
                            )
        elif isinstance(ml, dict):
            for sm in ml["stressmodels"].values():
                classkey = "stressmodel" if PASTAS_LEQ_022 else "class"
                if sm[classkey] in prec_evap_model:
                    stresses = [sm["prec"], sm["evap"]]
                elif sm[classkey] in ["WellModel"]:
                    stresses = sm["stress"]
                else:
                    stresses = sm["stress"] if PASTAS_LEQ_022 else [sm["stress"]]
                for s in stresses:
                    if str(s["name"]) not in self.stresses.index:
                        msg = (
                            f"Cannot add model because stress '{s['name']}' "
                            "is not contained in store."
                        )
                        raise LookupError(msg)
        else:
            raise TypeError("Expected pastas.Model or dict!")

    def _stored_series_to_json(
        self,
        libname: str,
        names: Optional[Union[list, str]] = None,
        squeeze: bool = True,
        progressbar: bool = False,
    ):
        """Write stored series to JSON.

        Parameters
        ----------
        libname : str
            library name
        names : Optional[Union[list, str]], optional
            names of series, by default None
        squeeze : bool, optional
            return single entry as json string instead
            of list, by default True
        progressbar : bool, optional
            show progressbar, by default False

        Returns
        -------
        files : list or str
            list of series converted to JSON string or single string
            if single entry is returned and squeeze is True
        """
        names = self._parse_names(names, libname=libname)
        files = []
        for n in tqdm(names, desc=libname) if progressbar else names:
            s = self._get_series(libname, n, progressbar=False)
            if isinstance(s, pd.Series):
                s = s.to_frame()
            try:
                sjson = s.to_json(orient="columns")
            except ValueError as e:
                msg = (
                    f"DatetimeIndex of '{n}' probably contains NaT "
                    "or duplicate timestamps!"
                )
                raise ValueError(msg) from e
            files.append(sjson)
        if len(files) == 1 and squeeze:
            return files[0]
        else:
            return files

    def _stored_metadata_to_json(
        self,
        libname: str,
        names: Optional[Union[list, str]] = None,
        squeeze: bool = True,
        progressbar: bool = False,
    ):
        """Write metadata from stored series to JSON.

        Parameters
        ----------
        libname : str
            library containing series
        names : Optional[Union[list, str]], optional
            names to parse, by default None
        squeeze : bool, optional
            return single entry as json string instead of list, by default True
        progressbar : bool, optional
            show progressbar, by default False

        Returns
        -------
        files : list or str
            list of json string
        """
        names = self._parse_names(names, libname=libname)
        files = []
        for n in tqdm(names, desc=libname) if progressbar else names:
            meta = self.get_metadata(libname, n, as_frame=False)
            meta_json = json.dumps(meta, cls=PastasEncoder, indent=4)
            files.append(meta_json)
        if len(files) == 1 and squeeze:
            return files[0]
        else:
            return files

    def _series_to_archive(
        self,
        archive,
        libname: str,
        names: Optional[Union[list, str]] = None,
        progressbar: bool = True,
    ):
        """Write DataFrame or Series to zipfile (internal method).

        Parameters
        ----------
        archive : zipfile.ZipFile
            reference to an archive to write data to
        libname : str
            name of the library to write to zipfile
        names : str or list of str, optional
            names of the time series to write to archive, by default None,
            which writes all time series to archive
        progressbar : bool, optional
            show progressbar, by default True
        """
        names = self._parse_names(names, libname=libname)
        for n in tqdm(names, desc=libname) if progressbar else names:
            sjson = self._stored_series_to_json(
                libname, names=n, progressbar=False, squeeze=True
            )
            meta_json = self._stored_metadata_to_json(
                libname, names=n, progressbar=False, squeeze=True
            )
            archive.writestr(f"{libname}/{n}.json", sjson)
            archive.writestr(f"{libname}/{n}_meta.json", meta_json)

    def _models_to_archive(self, archive, names=None, progressbar=True):
        """Write pastas.Model to zipfile (internal method).

        Parameters
        ----------
        archive : zipfile.ZipFile
            reference to an archive to write data to
        names : str or list of str, optional
            names of the models to write to archive, by default None,
            which writes all models to archive
        progressbar : bool, optional
            show progressbar, by default True
        """
        names = self._parse_names(names, libname="models")
        for n in tqdm(names, desc="models") if progressbar else names:
            m = self.get_models(n, return_dict=True)
            jsondict = json.dumps(m, cls=PastasEncoder, indent=4)
            archive.writestr(f"models/{n}.pas", jsondict)

    @staticmethod
    def _series_from_json(fjson: str, squeeze: bool = True):
        """Load time series from JSON.

        Parameters
        ----------
        fjson : str
            path to file
        squeeze : bool, optional
            squeeze time series object to obtain pandas Series

        Returns
        -------
        s : pd.DataFrame
            DataFrame containing time series
        """
        s = pd.read_json(fjson, orient="columns", precise_float=True, dtype=False)
        if not isinstance(s.index, pd.DatetimeIndex):
            s.index = pd.to_datetime(s.index, unit="ms")
        s = s.sort_index()  # needed for some reason ...
        if squeeze:
            return s.squeeze()
        return s

    @staticmethod
    def _metadata_from_json(fjson: str):
        """Load metadata dictionary from JSON.

        Parameters
        ----------
        fjson : str
            path to file

        Returns
        -------
        meta : dict
            dictionary containing metadata
        """
        with open(fjson, "r") as f:
            meta = json.load(f)
        return meta

    def _get_model_orphans(self):
        """Get models whose oseries no longer exist in database.

        Returns
        -------
        dict
            dictionary with oseries names as keys and lists of model names
            as values
        """
        d = {}
        for mlnam in tqdm(self.model_names, desc="Identifying model orphans"):
            mdict = self.get_models(mlnam, return_dict=True)
            onam = mdict["oseries"]["name"]
            if onam not in self.oseries_names:
                if onam in d:
                    d[onam] = d[onam].append(mlnam)
                else:
                    d[onam] = [mlnam]
        return d


class ModelAccessor:
    """Object for managing access to stored models.

    Provides dict-like access to models (i.e. PastaStore.models["model1"]), or allows
    adding models to the PastaStore using dict-like assignment (i.e.
    PastaStore.models["model1"] = ml), and it can serve as an iterator (i.e. [ml for ml
    in pstore.models]).
    """

    def __init__(self, conn):
        """Initialize model accessor.

        Parameters
        ----------
        conn : pastastore.*Connector type
            connector
        """
        self.conn = conn

    def __repr__(self):
        """Representation of the object is a list of modelnames."""
        return self.conn._modelnames_cache.__repr__()

    def __getitem__(self, name: str):
        """Get model from store with model name as key.

        Parameters
        ----------
        name : str
            name of the model
        """
        return self.conn.get_models(name)

    def __setitem__(self, name: str, ml: ps.Model):
        """Set item.

        Parameters
        ----------
        name : str
            name of the model
        ml : pastas.Model or dict
            model to add to the pastastore
        """
        ml.name = name
        self.conn.add_model(ml, overwrite=True)

    def __iter__(self):
        """Iterate over models.

        Yields
        ------
        ml : pastas.Model
            model
        """
        yield from self.conn.iter_models()

    def __len__(self):
        """No.

        of models
        """
        return self.conn.n_models

    def random(self):
        """Return a random model.

        Returns
        -------
        pastas.Model
            A random model object from the connection.
        """
        from random import choice

        return self.conn.get_models(choice(self.conn._modelnames_cache))



"""Module containing classes for connecting to different data stores."""

import json
import os
import warnings
from copy import deepcopy
from typing import Dict, Optional, Union

import pandas as pd
from pastas.io.pas import PastasEncoder, pastas_hook

from pastastore.base import BaseConnector, ConnectorUtil, ModelAccessor
from pastastore.util import _custom_warning

FrameorSeriesUnion = Union[pd.DataFrame, pd.Series]
warnings.showwarning = _custom_warning


class ArcticDBConnector(BaseConnector, ConnectorUtil):
    """ArcticDBConnector object using ArcticDB to store data."""

    conn_type = "arcticdb"

    def __init__(self, name: str, uri: str):
        """Create an ArcticDBConnector object using ArcticDB to store data.

        Parameters
        ----------
        name : str
            name of the database
        uri : str
            URI connection string (e.g. 'lmdb://<your path here>')
        """
        try:
            import arcticdb
        except ModuleNotFoundError as e:
            print("Please install arcticdb with `pip install arcticdb`!")
            raise e
        self.uri = uri
        self.name = name

        self.libs: dict = {}
        self.arc = arcticdb.Arctic(uri)
        self._initialize()
        self.models = ModelAccessor(self)
        # for older versions of PastaStore, if oseries_models library is empty
        # populate oseries - models database
        self._update_all_oseries_model_links()

    def _initialize(self) -> None:
        """Initialize the libraries (internal method)."""
        for libname in self._default_library_names:
            if self._library_name(libname) not in self.arc.list_libraries():
                self.arc.create_library(self._library_name(libname))
            else:
                print(
                    f"ArcticDBConnector: library "
                    f"'{self._library_name(libname)}'"
                    " already exists. Linking to existing library."
                )
            self.libs[libname] = self._get_library(libname)

    def _library_name(self, libname: str) -> str:
        """Get full library name according to ArcticDB (internal method)."""
        return ".".join([self.name, libname])

    def _get_library(self, libname: str):
        """Get ArcticDB library handle.

        Parameters
        ----------
        libname : str
            name of the library

        Returns
        -------
        lib : arcticdb.Library handle
            handle to the library
        """
        # get library handle
        lib = self.arc.get_library(self._library_name(libname))
        return lib

    def _add_item(
        self,
        libname: str,
        item: Union[FrameorSeriesUnion, Dict],
        name: str,
        metadata: Optional[Dict] = None,
        **_,
    ) -> None:
        """Add item to library (time series or model) (internal method).

        Parameters
        ----------
        libname : str
            name of the library
        item : Union[FrameorSeriesUnion, Dict]
            item to add, either time series or pastas.Model as dictionary
        name : str
            name of the item
        metadata : Optional[Dict], optional
            dictionary containing metadata, by default None
        """
        lib = self._get_library(libname)
        # only normalizable datatypes can be written with write, else use write_pickle
        # normalizable: Series, DataFrames, Numpy Arrays
        if isinstance(item, (dict, list)):
            lib.write_pickle(name, item, metadata=metadata)
        else:
            lib.write(name, item, metadata=metadata)

    def _get_item(self, libname: str, name: str) -> Union[FrameorSeriesUnion, Dict]:
        """Retrieve item from library (internal method).

        Parameters
        ----------
        libname : str
            name of the library
        name : str
            name of the item

        Returns
        -------
        item : Union[FrameorSeriesUnion, Dict]
            time series or model dictionary
        """
        lib = self._get_library(libname)
        return lib.read(name).data

    def _del_item(self, libname: str, name: str) -> None:
        """Delete items (series or models) (internal method).

        Parameters
        ----------
        libname : str
            name of library to delete item from
        name : str
            name of item to delete
        """
        lib = self._get_library(libname)
        lib.delete(name)

    def _get_metadata(self, libname: str, name: str) -> dict:
        """Retrieve metadata for an item (internal method).

        Parameters
        ----------
        libname : str
            name of the library
        name : str
            name of the item

        Returns
        -------
        dict
            dictionary containing metadata
        """
        lib = self._get_library(libname)
        return lib.read_metadata(name).metadata

    @property
    def oseries_names(self):
        """List of oseries names.

        Returns
        -------
        list
            list of oseries in library
        """
        return self._get_library("oseries").list_symbols()

    @property
    def stresses_names(self):
        """List of stresses names.

        Returns
        -------
        list
            list of stresses in library
        """
        return self._get_library("stresses").list_symbols()

    @property
    def model_names(self):
        """List of model names.

        Returns
        -------
        list
            list of models in library
        """
        return self._get_library("models").list_symbols()

    @property
    def oseries_with_models(self):
        """List of oseries with models."""
        return self._get_library("oseries_models").list_symbols()


class DictConnector(BaseConnector, ConnectorUtil):
    """DictConnector object that stores timeseries and models in dictionaries."""

    conn_type = "dict"

    def __init__(self, name: str = "pastas_db"):
        """Create DictConnector object that stores data in dictionaries.

        Parameters
        ----------
        name : str, optional
            user-specified name of the connector
        """
        self.name = name

        # create empty dictionaries for series and models
        for val in self._default_library_names:
            setattr(self, "lib_" + val, {})
        self.models = ModelAccessor(self)
        # for older versions of PastaStore, if oseries_models library is empty
        # populate oseries - models database
        self._update_all_oseries_model_links()

    def _get_library(self, libname: str):
        """Get reference to dictionary holding data.

        Parameters
        ----------
        libname : str
            name of the library

        Returns
        -------
        lib : dict
            library handle
        """
        return getattr(self, f"lib_{libname}")

    def _add_item(
        self,
        libname: str,
        item: Union[FrameorSeriesUnion, Dict],
        name: str,
        metadata: Optional[Dict] = None,
        **_,
    ) -> None:
        """Add item (time series or models) (internal method).

        Parameters
        ----------
        libname : str
            name of library
        item : FrameorSeriesUnion
            pandas.Series or pandas.DataFrame containing data
        name : str
            name of the item
        metadata : dict, optional
            dictionary containing metadata, by default None
        """
        lib = self._get_library(libname)
        if libname in ["models", "oseries_models"]:
            lib[name] = item
        else:
            lib[name] = (metadata, item)

    def _get_item(self, libname: str, name: str) -> Union[FrameorSeriesUnion, Dict]:
        """Retrieve item from database (internal method).

        Parameters
        ----------
        libname : str
            name of the library
        name : str
            name of the item

        Returns
        -------
        item : Union[FrameorSeriesUnion, Dict]
            time series or model dictionary
        """
        lib = self._get_library(libname)
        if libname in ["models", "oseries_models"]:
            item = deepcopy(lib[name])
        else:
            item = deepcopy(lib[name][1])
        return item

    def _del_item(self, libname: str, name: str) -> None:
        """Delete items (series or models) (internal method).

        Parameters
        ----------
        libname : str
            name of library to delete item from
        name : str
            name of item to delete
        """
        lib = self._get_library(libname)
        _ = lib.pop(name)

    def _get_metadata(self, libname: str, name: str) -> dict:
        """Read metadata (internal method).

        Parameters
        ----------
        libname : str
            name of the library the series are in ("oseries" or "stresses")
        name : str
            name of item to load metadata for

        Returns
        -------
        imeta : dict
            dictionary containing metadata
        """
        lib = self._get_library(libname)
        imeta = deepcopy(lib[name][0])
        return imeta

    @property
    def oseries_names(self):
        """List of oseries names."""
        lib = self._get_library("oseries")
        return list(lib.keys())

    @property
    def stresses_names(self):
        """List of stresses names."""
        lib = self._get_library("stresses")
        return list(lib.keys())

    @property
    def model_names(self):
        """List of model names."""
        lib = self._get_library("models")
        return list(lib.keys())

    @property
    def oseries_with_models(self):
        """List of oseries with models."""
        lib = self._get_library("oseries_models")
        return list(lib.keys())


class PasConnector(BaseConnector, ConnectorUtil):
    """PasConnector object that stores time series and models as JSON files on disk."""

    conn_type = "pas"

    def __init__(self, name: str, path: str):
        """Create PasConnector object that stores data as JSON files on disk.

        Uses Pastas export format (pas-files) to store files.

        Parameters
        ----------
        name : str
            user-specified name of the connector, this will be the name of the
            directory in which the data will be stored
        path : str
            path to directory for storing the data
        """
        self.name = name
        self.path = os.path.abspath(os.path.join(path, self.name))
        self.relpath = os.path.relpath(self.path)
        self._initialize()
        self.models = ModelAccessor(self)
        # for older versions of PastaStore, if oseries_models library is empty
        # populate oseries_models library
        self._update_all_oseries_model_links()

    def _initialize(self) -> None:
        """Initialize the libraries (internal method)."""
        for val in self._default_library_names:
            libdir = os.path.join(self.path, val)
            if not os.path.exists(libdir):
                print(f"PasConnector: library '{val}' created in '{libdir}'")
                os.makedirs(libdir)
            else:
                print(
                    f"PasConnector: library '{val}' already exists. "
                    f"Linking to existing directory: '{libdir}'"
                )
            setattr(self, f"lib_{val}", os.path.join(self.path, val))

    def _get_library(self, libname: str):
        """Get path to directory holding data.

        Parameters
        ----------
        libname : str
            name of the library

        Returns
        -------
        lib : str
            path to library
        """
        return getattr(self, "lib_" + libname)

    def _add_item(
        self,
        libname: str,
        item: Union[FrameorSeriesUnion, Dict],
        name: str,
        metadata: Optional[Dict] = None,
        **_,
    ) -> None:
        """Add item (time series or models) (internal method).

        Parameters
        ----------
        libname : str
            name of library
        item : FrameorSeriesUnion
            pandas.Series or pandas.DataFrame containing data
        name : str
            name of the item
        metadata : dict, optional
            dictionary containing metadata, by default None
        """
        lib = self._get_library(libname)

        # time series
        if isinstance(item, pd.Series):
            item = item.to_frame()
        if isinstance(item, pd.DataFrame):
            sjson = item.to_json(orient="columns")
            fname = os.path.join(lib, f"{name}.pas")
            with open(fname, "w") as f:
                f.write(sjson)
            if metadata is not None:
                mjson = json.dumps(metadata, cls=PastasEncoder, indent=4)
                fname_meta = os.path.join(lib, f"{name}_meta.pas")
                with open(fname_meta, "w") as m:
                    m.write(mjson)
        # pastas model dict
        elif isinstance(item, dict):
            jsondict = json.dumps(item, cls=PastasEncoder, indent=4)
            fmodel = os.path.join(lib, f"{name}.pas")
            with open(fmodel, "w") as fm:
                fm.write(jsondict)
        # oseries_models list
        elif isinstance(item, list):
            jsondict = json.dumps(item)
            fname = os.path.join(lib, f"{name}.pas")
            with open(fname, "w") as fm:
                fm.write(jsondict)

    def _get_item(self, libname: str, name: str) -> Union[FrameorSeriesUnion, Dict]:
        """Retrieve item (internal method).

        Parameters
        ----------
        libname : str
            name of the library
        name : str
            name of the item

        Returns
        -------
        item : Union[FrameorSeriesUnion, Dict]
            time series or model dictionary
        """
        lib = self._get_library(libname)
        fjson = os.path.join(lib, f"{name}.pas")
        if not os.path.exists(fjson):
            msg = f"Item '{name}' not in '{libname}' library."
            raise FileNotFoundError(msg)
        # model
        if libname == "models":
            with open(fjson, "r") as ml_json:
                item = json.load(ml_json, object_hook=pastas_hook)
        # list of models per oseries
        elif libname == "oseries_models":
            with open(fjson, "r") as f:
                item = json.load(f)
        # time series
        else:
            item = self._series_from_json(fjson)
        return item

    def _del_item(self, libname: str, name: str) -> None:
        """Delete items (series or models) (internal method).

        Parameters
        ----------
        libname : str
            name of library to delete item from
        name : str
            name of item to delete
        """
        lib = self._get_library(libname)
        os.remove(os.path.join(lib, f"{name}.pas"))
        # remove metadata for time series
        if libname != "models":
            try:
                os.remove(os.path.join(lib, f"{name}_meta.pas"))
            except FileNotFoundError:
                # Nothing to delete
                pass

    def _get_metadata(self, libname: str, name: str) -> dict:
        """Read metadata (internal method).

        Parameters
        ----------
        libname : str
            name of the library the series are in ("oseries" or "stresses")
        name : str
            name of item to load metadata for

        Returns
        -------
        imeta : dict
            dictionary containing metadata
        """
        lib = self._get_library(libname)
        mjson = os.path.join(lib, f"{name}_meta.pas")
        if os.path.isfile(mjson):
            imeta = self._metadata_from_json(mjson)
        else:
            imeta = {}
        return imeta

    @property
    def oseries_names(self):
        """List of oseries names."""
        lib = self._get_library("oseries")
        return [
            i[:-4]
            for i in os.listdir(lib)
            if i.endswith(".pas")
            if not i.endswith("_meta.pas")
        ]

    @property
    def stresses_names(self):
        """List of stresses names."""
        lib = self._get_library("stresses")
        return [
            i[:-4]
            for i in os.listdir(lib)
            if i.endswith(".pas")
            if not i.endswith("_meta.pas")
        ]

    @property
    def model_names(self):
        """List of model names."""
        lib = self._get_library("models")
        return [i[:-4] for i in os.listdir(lib) if i.endswith(".pas")]

    @property
    def oseries_with_models(self):
        """List of oseries with models."""
        lib = self._get_library("oseries_models")
        return [i[:-4] for i in os.listdir(lib) if i.endswith(".pas")]



"""Module containing example dataset."""

import os

import pandas as pd
from hydropandas import Obs, ObsCollection

try:
    from pastas.timeseries_utils import timestep_weighted_resample
except ModuleNotFoundError:
    from pastas.utils import timestep_weighted_resample

import pastastore as pst
from pastastore.base import BaseConnector


def example_pastastore(conn="DictConnector"):
    """Get example PastaStore.

    Parameters
    ----------
    conn : str or Connector, optional
        name of Connector type, by default "DictConnector", which
        initializes a default Connector. If an Connector instance is passed,
        use that Connector.

    Returns
    -------
    pstore : pastastore.PastaStore
        PastaStore containing example dataset
    """
    # check it test dataset is available
    datadir = os.path.join(os.path.dirname(__file__), "../tests/data")
    if not os.path.exists(datadir):
        raise FileNotFoundError(
            "Test datasets not available! Clone repository from GitHub."
        )

    # initialize default connector if conn is str
    if not isinstance(conn, BaseConnector):
        conn = _default_connector(conn)

    # initialize PastaStore
    pstore = pst.PastaStore(conn, "example")

    # add data

    # oseries 1
    o = pd.read_csv(os.path.join(datadir, "obs.csv"), index_col=0, parse_dates=True)
    pstore.add_oseries(o, "oseries1", metadata={"x": 165000, "y": 424000})
    # oseries 2
    o = pd.read_csv(
        os.path.join(datadir, "head_nb1.csv"), index_col=0, parse_dates=True
    )
    pstore.add_oseries(o, "oseries2", metadata={"x": 164000, "y": 423000})

    # oseries 3
    o = pd.read_csv(os.path.join(datadir, "gw_obs.csv"), index_col=0, parse_dates=True)
    pstore.add_oseries(o, "oseries3", metadata={"x": 165554, "y": 422685})

    # prec 1
    s = pd.read_csv(os.path.join(datadir, "rain.csv"), index_col=0, parse_dates=True)
    pstore.add_stress(s, "prec1", kind="prec", metadata={"x": 165050, "y": 424050})

    # prec 2
    s = pd.read_csv(
        os.path.join(datadir, "rain_nb1.csv"), index_col=0, parse_dates=True
    )
    pstore.add_stress(s, "prec2", kind="prec", metadata={"x": 164010, "y": 423000})

    # evap 1
    s = pd.read_csv(os.path.join(datadir, "evap.csv"), index_col=0, parse_dates=True)
    pstore.add_stress(s, "evap1", kind="evap", metadata={"x": 164500, "y": 424000})

    # evap 2
    s = pd.read_csv(
        os.path.join(datadir, "evap_nb1.csv"), index_col=0, parse_dates=True
    )
    pstore.add_stress(s, "evap2", kind="evap", metadata={"x": 164000, "y": 423030})

    # well 1
    s = pd.read_csv(os.path.join(datadir, "well.csv"), index_col=0, parse_dates=True)
    s = timestep_weighted_resample(s, pd.date_range(s.index[0], s.index[-1], freq="D"))
    pstore.add_stress(s, "well1", kind="well", metadata={"x": 164691, "y": 423579})

    # river notebook data (nb5)
    oseries = pd.read_csv(
        os.path.join(datadir, "nb5_head.csv"), parse_dates=True, index_col=0
    ).squeeze("columns")
    pstore.add_oseries(oseries, "head_nb5", metadata={"x": 200_000, "y": 450_000.0})

    rain = pd.read_csv(
        os.path.join(datadir, "nb5_prec.csv"), parse_dates=True, index_col=0
    ).squeeze("columns")
    pstore.add_stress(
        rain, "prec_nb5", kind="prec", metadata={"x": 200_000, "y": 450_000.0}
    )
    evap = pd.read_csv(
        os.path.join(datadir, "nb5_evap.csv"), parse_dates=True, index_col=0
    ).squeeze("columns")
    pstore.add_stress(
        evap, "evap_nb5", kind="evap", metadata={"x": 200_000, "y": 450_000.0}
    )
    waterlevel = pd.read_csv(
        os.path.join(datadir, "nb5_riv.csv"), parse_dates=True, index_col=0
    ).squeeze("columns")
    pstore.add_stress(
        waterlevel,
        "riv_nb5",
        kind="riv",
        metadata={"x": 200_000, "y": 450_000.0},
    )
    # TODO: temporary fix for older version of hydropandas that does not
    # read Menyanthes time series names correctly.
    # multiwell notebook data
    fname = os.path.join(datadir, "MenyanthesTest.men")
    # meny = ps.read.MenyData(fname)
    meny = ObsCollection.from_menyanthes(fname, Obs)

    oseries = meny.loc["Obsevation well", "obs"]
    ometa = {
        "x": oseries.meta["x"],
        "y": oseries.meta["y"],
    }
    pstore.add_oseries(oseries.dropna(), "head_mw", metadata=ometa)

    prec = meny.loc["Precipitation", "obs"]
    prec.index = prec.index.round("D")
    prec.name = "prec"
    pmeta = {
        "x": prec.meta["x"],
        "y": prec.meta["y"],
    }
    pstore.add_stress(prec, "prec_mw", kind="prec", metadata=pmeta)
    evap = meny.loc["Evaporation", "obs"]
    evap.index = evap.index.round("D")
    evap.name = "evap"
    emeta = {
        "x": evap.meta["x"],
        "y": evap.meta["y"],
    }
    pstore.add_stress(evap, "evap_mw", kind="evap", metadata=emeta)

    pressure = meny.loc["Air Pressure", "obs"]
    pressure.index = pressure.index.round("D")
    pressure.name = "pressure"
    pres_meta = {
        "x": pressure.meta["x"],
        "y": pressure.meta["y"],
    }
    pstore.add_stress(pressure, "pressure_mw", kind="pressure", metadata=pres_meta)

    extraction_names = [
        "Extraction 1",
        "Extraction 2",
        "Extraction 3",
        "Extraction 4",
    ]
    for extr in extraction_names:
        ts = meny.loc[extr, "obs"]
        wmeta = {"x": ts.meta["x"], "y": ts.meta["y"]}
        # replace spaces in names for Pastas
        name = extr.replace(" ", "_").lower()
        # resample to daily timestep
        ts_d = timestep_weighted_resample(
            ts, pd.date_range(ts.index[0], ts.index[-1], freq="D")
        )
        pstore.add_stress(ts_d, name, kind="well", metadata=wmeta)

    return pstore


def _default_connector(conntype: str):
    """Get default connector based on name.

    Parameters
    ----------
    conntype : str
        name of connector (DictConnector, PasConnector, ArcticDBConnector)

    Returns
    -------
    conn : Connector
        default Connector based on type.
    """
    Conn = getattr(pst, conntype)
    if Conn.conn_type == "arcticdb":
        uri = "lmdb://./arctic_db"
        conn = Conn("my_db", uri)
    elif Conn.conn_type == "dict":
        conn = Conn("my_db")
    elif Conn.conn_type == "pas":
        conn = Conn("my_db", "./pas_db")
    else:
        raise ValueError(f"Unrecognized connector type! '{conntype}'")
    return conn



# ruff: noqa: F401 D104
from pastastore import connectors, styling, util
from pastastore.connectors import (
    ArcticDBConnector,
    DictConnector,
    PasConnector,
)
from pastastore.store import PastaStore
from pastastore.version import __version__, show_versions



"""Module containing dataframe styling functions."""

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np


def float_styler(val, norm, cmap=None):
    """Style float values in DataFrame.

    Parameters
    ----------
    val : float
        value in cell
    norm : matplotlib.colors.Normalize
        normalizer to map values to range(0, 1)
    cmap : colormap, optional
        colormap to use, by default None, which uses RdYlBu

    Returns
    -------
    str
        css value pairs for styling dataframe

    Usage
    -----
    Given some dataframe

    >>> df.map(float_styler, subset=["some column"], norm=norm, cmap=cmap)
    """
    if cmap is None:
        cmap = plt.get_cmap("RdYlBu")
    bg = cmap(norm(val))
    color = mpl.colors.rgb2hex(bg)
    c = "White" if np.mean(bg[:3]) < 0.4 else "Black"
    return f"background-color: {color}; color: {c}"


def boolean_styler(b):
    """Style boolean values in DataFrame.

    Parameters
    ----------
    b : bool
        value in cell

    Returns
    -------
    str
        css value pairs for styling dataframe

    Usage
    -----
    Given some dataframe

    >>> df.map(boolean_styler, subset=["some column"])
    """
    if b:
        return (
            f"background-color: {mpl.colors.rgb2hex((231/255, 255/255, 239/255))}; "
            "color: darkgreen"
        )
    else:
        return (
            f"background-color: {mpl.colors.rgb2hex((255/255, 238/255, 238/255))}; "
            "color: darkred"
        )



# ruff: noqa: D100 D103
import os

import numpy as np
import pandas as pd
import pastas as ps
import pytest
from numpy import allclose
from packaging.version import parse
from pytest_dependency import depends

import pastastore as pst


@pytest.mark.dependency
def test_iter_oseries(pstore):
    _ = list(pstore.iter_oseries())


@pytest.mark.dependency
def test_iter_stresses(pstore):
    _ = list(pstore.iter_stresses())


@pytest.mark.dependency
def test_get_tmintmax(pstore):
    ostt = pstore.get_tmin_tmax("oseries")
    assert ostt.at["oseries1", "tmin"] == pd.Timestamp("2010-01-14")
    sttt = pstore.get_tmin_tmax("stresses")
    assert sttt.at["evap2", "tmax"] == pd.Timestamp("2016-11-22")
    ml = pstore.create_model("oseries1")
    ml.solve(report=False)
    pstore.conn.add_model(ml)
    mltt = pstore.get_tmin_tmax("models")
    assert mltt.at["oseries1", "tmax"] == pd.Timestamp("2015-06-28")
    pstore.del_model("oseries1")


@pytest.mark.dependency
def test_search(pstore):
    results = pstore.search("oseries", "OSER", case_sensitive=False)
    assert len(results) == 3
    assert len(set(results) - {"oseries1", "oseries2", "oseries3"}) == 0


@pytest.mark.dependency
def test_create_model(pstore):
    _ = pstore.create_model("oseries1")


@pytest.mark.dependency
def test_properties(pstore):
    pstore.add_oseries(pd.Series(dtype=np.float64), "deleteme", validate=False)
    pstore.add_stress(
        pd.Series(dtype=np.float64), "deleteme", kind="useless", validate=False
    )

    _ = pstore.oseries
    _ = pstore.stresses
    _ = pstore.models

    try:
        assert pstore.n_oseries == pstore.conn.n_oseries
        assert pstore.n_stresses == pstore.conn.n_stresses
    finally:
        pstore.del_oseries("deleteme")
        pstore.del_stress("deleteme")


@pytest.mark.dependency
def test_store_model(request, pstore):
    depends(request, [f"test_create_model[{pstore.type}]"])
    ml = pstore.create_model("oseries1")
    pstore.conn.add_model(ml)


@pytest.mark.dependency
def test_model_accessor(request, pstore):
    depends(request, [f"test_store_model[{pstore.type}]"])
    # repr
    pstore.models.__repr__()
    # getter
    ml = pstore.models["oseries1"]
    # setter
    pstore.models["oseries1_2"] = ml
    # iter
    mnames = [ml.name for ml in pstore.models]
    try:
        assert len(mnames) == 2
        assert mnames[0] in ["oseries1", "oseries1_2"]
        assert mnames[1] in ["oseries1", "oseries1_2"]
    finally:
        pstore.del_models("oseries1_2")


@pytest.mark.dependency
def test_oseries_model_accessor(request, pstore):
    depends(request, [f"test_store_model[{pstore.type}]"])
    # repr
    pstore.oseries_models.__repr__()
    # get model names
    ml = pstore.models["oseries1"]
    ml_list1 = pstore.oseries_models["oseries1"]
    assert len(ml_list1) == 1

    # add model
    pstore.models["oseries1_2"] = ml
    ml_list2 = pstore.oseries_models["oseries1"]
    assert len(ml_list2) == 2

    # delete model
    pstore.del_models("oseries1_2")
    ml_list3 = pstore.oseries_models["oseries1"]
    assert len(ml_list3) == 1


@pytest.mark.dependency
def test_store_model_missing_series(request, pstore):
    depends(
        request,
        [
            f"test_create_model[{pstore.type}]",
            f"test_store_model[{pstore.type}]",
        ],
    )
    ml = pstore.create_model("oseries1")
    o = pstore.get_oseries("oseries1")
    meta = pstore.get_metadata("oseries", "oseries1", as_frame=False)
    pstore.del_models("oseries1")
    pstore.del_oseries("oseries1")
    try:
        pstore.add_model(ml)
    except LookupError:
        pstore.add_oseries(o, "oseries1", metadata=meta)
        pstore.add_model(ml)


@pytest.mark.dependency
def test_get_model(request, pstore):
    depends(
        request,
        [
            f"test_create_model[{pstore.type}]",
            f"test_store_model[{pstore.type}]",
            f"test_store_model_missing_series[{pstore.type}]",
        ],
    )
    _ = pstore.conn.get_models("oseries1")


@pytest.mark.dependency
def test_del_model(request, pstore):
    depends(
        request,
        [
            f"test_create_model[{pstore.type}]",
            f"test_store_model[{pstore.type}]",
            f"test_store_model_missing_series[{pstore.type}]",
            f"test_get_model[{pstore.type}]",
        ],
    )
    pstore.conn.del_models("oseries1")


@pytest.mark.dependency
def test_create_models(pstore):
    _ = pstore.create_models_bulk(
        ["oseries1", "oseries2"], store=True, progressbar=False
    )
    _ = pstore.conn.models


@pytest.mark.dependency
def test_get_parameters(request, pstore):
    depends(request, [f"test_create_models[{pstore.type}]"])
    p = pstore.get_parameters(progressbar=False, param_value="initial")
    assert p.index.size == 2
    assert p.isna().sum().sum() == 0


@pytest.mark.dependency
def test_get_signatures(request, pstore):
    depends(request, [f"test_create_models[{pstore.type}]"])
    s = pstore.get_signatures(progressbar=False)
    assert s.shape[1] == len(ps.stats.signatures.__all__)


@pytest.mark.dependency
def test_iter_models(request, pstore):
    depends(request, [f"test_create_models[{pstore.type}]"])
    _ = list(pstore.iter_models())


@pytest.mark.dependency
def test_solve_models_and_get_stats(request, pstore):
    depends(request, [f"test_create_models[{pstore.type}]"])
    _ = pstore.solve_models(
        ignore_solve_errors=False, progressbar=False, store_result=True
    )
    stats = pstore.get_statistics(["evp", "aic"], progressbar=False)
    assert stats.index.size == 2


@pytest.mark.dependency
def test_apply(request, pstore):
    depends(request, [f"test_solve_models_and_get_stats[{pstore.type}]"])

    def func(ml):
        return ml.parameters.loc["recharge_A", "optimal"]

    result = pstore.apply("models", func)
    assert len(result) == 2


@pytest.mark.dependency
def test_save_and_load_model(request, pstore):
    ml = pstore.create_model("oseries1")
    ml.solve()
    evp_ml = ml.stats.evp()
    pstore.add_model(ml, overwrite=True)
    ml2 = pstore.get_models(ml.name)
    evp_ml2 = ml2.stats.evp()
    assert allclose(evp_ml, evp_ml2)
    assert pst.util.compare_models(ml, ml2)


def test_update_ts_settings(request, pstore):
    pstore.set_check_model_series_values(False)

    o = pstore.get_oseries("oseries2")
    ml = ps.Model(o.loc[:"2013"], name="ml_oseries2")

    pnam = pstore.get_nearest_stresses("oseries2", kind="prec").iloc[0]
    p = pstore.get_stresses(pnam)
    enam = pstore.get_nearest_stresses("oseries2", kind="evap").iloc[0]
    e = pstore.get_stresses(enam)
    rm = ps.RechargeModel(p.loc[:"2013"], e.loc[:"2013"])
    ml.add_stressmodel(rm)
    tmax = p.index.intersection(e.index).max()
    ml.solve()

    p2 = pstore.get_stresses("prec1")
    sm = ps.StressModel(p2.loc[:"2013"], ps.Exponential(), "prec")
    ml.add_stressmodel(sm)
    pstore.add_model(ml)

    ml2 = pstore.get_models(ml.name, update_ts_settings=True)

    try:
        assert ml2.oseries.settings["tmax"] == o.index[-1]
        assert ml2.stressmodels["recharge"].prec.settings["tmax"] == tmax
        assert ml2.stressmodels["recharge"].evap.settings["tmax"] == tmax
        assert ml2.stressmodels["prec"].stress[0].settings["tmax"] == p2.index[-1]
    except AssertionError:
        pstore.del_models("ml_oseries2")
        pstore.set_check_model_series_values(True)
        raise


# @pytest.mark.dependency()
# def test_model_results(request, pstore):
#     depends(request, [f"test_create_models[{pstore.type}]",
#                       f"test_solve_models[{pstore.type}]"])
#     pstore.model_results(["oseries1", "oseries2"], progressbar=False)


def test_oseries_distances(pstore):
    _ = pstore.get_nearest_oseries()


def test_repr(pstore):
    pstore.__repr__()


def test_copy_dbase(pstore):
    conn2 = pst.DictConnector("destination")
    pst.util.copy_database(pstore.conn, conn2, overwrite=False, progressbar=True)


def test_to_from_zip(pstore):
    if pstore.type == "arcticdb" and parse(ps.__version__) < parse("1.1.0"):
        pytest.xfail("model datetime objects not supported")
    zipname = f"test_{pstore.type}.zip"
    pstore.to_zip(zipname, progressbar=False, overwrite=True)
    conn = pst.DictConnector("test")
    try:
        store = pst.PastaStore.from_zip(zipname, conn)
        assert not store.oseries.empty
    finally:
        os.remove(zipname)


def test_example_pastastore():
    from pastastore.datasets import example_pastastore

    _ = example_pastastore()


def test_validate_names():
    from pastastore.util import validate_names

    assert validate_names(s="(test)") == "test"
    assert validate_names(d={"(test)": 2})["test"] == 2


def test_meta_with_name(pstore):
    s = pd.Series(
        index=pd.date_range("2020-01-01", periods=10, freq="D"),
        data=np.arange(10),
        dtype=float,
    )
    smeta = {"name": "not_what_i_want"}
    pstore.add_stress(s, "what_i_want", kind="special", metadata=smeta)
    assert "what_i_want" in pstore.stresses.index, "This is not right."
    pstore.del_stress("what_i_want")



# ruff: noqa: D100 D103
import pytest
from pandas import Timestamp

import pastastore as pst


@pytest.mark.pastas150
def test_hpd_download_from_bro():
    from pastastore.extensions import activate_hydropandas_extension

    activate_hydropandas_extension()
    pstore = pst.PastaStore()
    pstore.hpd.download_bro_gmw(
        extent=(117850, 118180, 439550, 439900), tmin="2022-01-01", tmax="2022-01-02"
    )
    assert pstore.n_oseries == 3


@pytest.mark.pastas150
def test_hpd_download_precipitation_from_knmi():
    from pastastore.extensions import activate_hydropandas_extension

    activate_hydropandas_extension()
    pstore = pst.PastaStore()
    pstore.hpd.download_knmi_precipitation(
        stns=[260], tmin="2022-01-01", tmax="2022-01-31"
    )
    assert pstore.n_stresses == 1


@pytest.mark.pastas150
def test_hpd_download_evaporation_from_knmi():
    from pastastore.extensions import activate_hydropandas_extension

    activate_hydropandas_extension()
    pstore = pst.PastaStore()
    pstore.hpd.download_knmi_evaporation(
        stns=[260], tmin="2022-01-01", tmax="2022-01-31"
    )
    assert pstore.n_stresses == 1


@pytest.mark.pastas150
def test_update_oseries():
    from pastastore.extensions import activate_hydropandas_extension

    activate_hydropandas_extension()

    pstore = pst.PastaStore.from_zip("tests/data/test_hpd_update.zip")
    pstore.hpd.update_bro_gmw(tmax="2024-01-31")
    tmintmax = pstore.get_tmin_tmax("oseries")
    assert tmintmax.loc["GMW000000036319_1", "tmax"] >= Timestamp("2024-01-30")
    assert tmintmax.loc["GMW000000036327_1", "tmax"] >= Timestamp("2024-01-20")


@pytest.mark.pastas150
def test_update_stresses():
    from pastastore.extensions import activate_hydropandas_extension

    activate_hydropandas_extension()

    pstore = pst.PastaStore.from_zip("tests/data/test_hpd_update.zip")
    pstore.hpd.update_knmi_meteo(tmax="2024-01-31")
    tmintmax = pstore.get_tmin_tmax("stresses")
    assert (tmintmax["tmax"] >= Timestamp("2024-01-31")).all()
